import numpy as np
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from RTF.nrf.Neural_Decision_Tree import *

dataset_length = 100
D = 2

X = np.random.randn(dataset_length, D)*0.1
X[0:dataset_length//2, 0] += 0.1
X[0:dataset_length//2, 0] += 0.2
Y = np.ones(dataset_length)
Y[0:dataset_length//2] *= 0

X_test = np.random.randn(dataset_length, D)*0.1
X_test[0:dataset_length//2, 0] += 0.1
X_test[0:dataset_length//2, 0] += 0.2
Y_test = np.ones(dataset_length)
Y_test[0:dataset_length//2] *= 0
# Train a Tree
clf = DecisionTreeClassifier(max_depth=5)
clf = clf.fit(X, Y)

a = NDTClassifier(D=2, gammas=[100, 1], tree_id=0)
a.compute_matrices_and_biases(clf)
a.to_keras("binary_crossentropy", dropouts=[0., 0., 0.])
print("FLOPs before:", a.count_ops)
a.fit(X, Y, verbose=1)
print("FLOPs after:", a.count_ops)

dataset_length = 100
D = 2

X = np.random.randn(dataset_length, D)*0.1
Y = X[:,0]**2+3*X[:, 1]+2

X_test = np.random.randn(dataset_length, D)*0.1
Y_test = X[:,0]**2+3*X[:, 1]+2
# Train a Tree
clf = DecisionTreeRegressor(max_depth=5)
clf = clf.fit(X, Y)

b = NDTRegressor(D=2, gammas=[100, 1], tree_id=0)
b.compute_matrices_and_biases(clf)
b.to_keras('mean_squared_error', dropouts=[0., 0., 0.],
            kernel_regularizer=[None, None, None])
print("FLOPs before:", b.count_ops)
b.fit(X, Y, verbose=1)
print("FLOPs after:", b.count_ops)


# -*- coding: utf-8 -*-
"""
Neural Random Forest Implementation
"""

# import sys

# sys.path.append("../forest")

from .Neural_Decision_Tree import ndt, NDTClassifier, NDTRegressor

# from ..forest.Forest import Forest
from ..forest.DT_grouping import *
from ..forest.Subforest import *

import numpy as np
import pandas as pd
from keras import optimizers
# from keras.callbacks import EarlyStopping, ModelCheckpoint
from keras.utils import plot_model
from keras.layers import Concatenate, Dense, Input
from keras.models import Model
from keras.regularizers import l1  # , l2
# from keras.utils import to_categorical
from sklearn.ensemble import RandomForestClassifier
# from sklearn.metrics import accuracy_score
# from sklearn.preprocessing import LabelEncoder
# from sklearn.tree import DecisionTreeClassifier, export_graphviz

# sys.path.append("../utils")
# from common_functions import (find_parent, get_list_split_phi,
# 							  get_list_split_phi_forest,
# 							  get_parents_nodes_leaves_dic, leaves_id)


class nrf_fully_connected(ndt):
    def __init__(self, D, gammas=[10, 1], random_forest_id=None, sigma=0,
                 gamma_activation=True, is_classifier=False):
        """
        Creates and neural random forest

        Args:
                D (int): Dataset dimensionality (nb. features)
                gammas (list): Metaparameter for each layer of the neural decision tree (slope of the tanh function).
                               High gamma -> behavior of NN is closer to tree (and also harder to change).
                random_forest_id (str or int): id for the neural random forest
                sigma (float): STD for the initial weights
        """
        ndt.__init__(self, D=D, gammas=gammas, tree_id=random_forest_id,
                     sigma=sigma, gamma_activation=gamma_activation,
                     is_classifier=is_classifier)

    def compute_matrices_and_biases(self, random_forest):
        """
        Computes the weights and biases of the NRF.
        One huge NN gathering all the trees is created.
        """
        self.rf = random_forest
        if self.is_classifier:
            self.classes = self.rf.classes_
        self.ndts = []
        # Compute weights and biases of individual trees
        for i, dtree in enumerate(random_forest.estimators_):
            self.ndts.append(ndt(D=self.D, gammas=self.gammas, tree_id=i,
                                 sigma=self.sigma, gamma_activation=self.use_gamma_activation,
                                 is_classifier=self.is_classifier))
            self.ndts[i].compute_matrices_and_biases(dtree)
        # Concatenate input to nodes W
        self.W_in_nodes = pd.concat([t.W_in_nodes for t in self.ndts], axis=1)
        # Concatenate input to nodes b
        self.b_nodes = pd.concat([t.b_nodes for t in self.ndts], axis=0)
        # Concatenate nodes to leaves W
        self.W_nodes_leaves = pd.concat([t.W_nodes_leaves for t in self.ndts], axis=1, sort=False)
        self.W_nodes_leaves = self.W_nodes_leaves.fillna(0)
        # Concatenate nodes to leaves b
        self.b_leaves = pd.concat([t.b_leaves for t in self.ndts], axis=0)
        # Concatenate leaves to out W
        self.W_leaves_out = pd.concat([t.W_leaves_out for t in self.ndts], axis=0)
        self.W_leaves_out = self.W_leaves_out/self.rf.n_estimators
        # Concatenate leaves to out b
        self.b_out = pd.concat([t.b_out for t in self.ndts], axis=0)
        self.b_out = self.b_out.groupby(self.b_out.index).mean()
        # Set other parameters
        self.N = self.b_nodes.size
        self.L = self.b_leaves.size
        self.C = self.b_out.size
        # print("self.C:", self.C)
        # print("n_classes", len(self.classes))


class NRFfullClassifier(nrf_fully_connected, NDTClassifier):
    def __init__(self, D, gammas=[10, 1], random_forest_id=None, sigma=0,
                 gamma_activation=True):
        super().__init__(D=D, gammas=gammas, random_forest_id=random_forest_id,
                         sigma=sigma, gamma_activation=gamma_activation,
                         is_classifier=True)


class NRFfullRegressor(nrf_fully_connected, NDTRegressor):
    def __init__(self, D, gammas=[10, 1], random_forest_id=None, sigma=0,
                 gamma_activation=True):
        super().__init__(D=D, gammas=gammas, random_forest_id=random_forest_id,
                         sigma=sigma, gamma_activation=gamma_activation,
                         is_classifier=False)


class nrf_independent_ndt(ndt):
    def __init__(self, D, gammas=[100, 1], sigma=0, gamma_activation=True,
                 is_classifier=False):
        """
        Creates a NRF were each tree is independent of each other.
        The last decision is a weighted voting of each tree

        Args:
        D (int): dimensionality of the dataset (nb. features)
        gammas (list): Metaparameter for each layer of the neural decision tree
        (slope of the tanh function).
        High gamma -> behavior of NN is closer to tree (and also harder to change).
        """
        ndt.__init__(self, D=D, gammas=gammas, sigma=sigma, gamma_activation=gamma_activation,
                     is_classifier=is_classifier)

    def compute_matrices_and_biases(self, random_forest):
        """
        Computes the weights and biases of the NRF.
        Each DT from the RF is mapped to a NN, and a last layer is created for
        the weighted voting.

        Args:
        random_forest (sklearn.ensemble.RandomForestClassifier): Sklearn random forest
        """
        self.rf = random_forest
        if self.is_classifier:
            self.classes = self.rf.classes_
        self.ndts = []
        for i, dtree in enumerate(random_forest.estimators_):
            self.ndts.append(ndt(D=self.D, gammas=self.gammas, tree_id=i,
                                 sigma=self.sigma, gamma_activation=self.use_gamma_activation,
                                 is_classifier=self.is_classifier))
            self.ndts[i].compute_matrices_and_biases(dtree)
        # # Define averaging layer
        # if self.is_classifier:
        # 	self.W_outputs_to_output = pd.DataFrame(np.concatenate([np.eye(tree.C)
        # 															for tree in self.ndts],
        # 															axis=0),
        # 											index=[str(t.tree_id)+"_"+str(t_class)
        # 												for t in self.ndts
        # 												for t_class in t.classes],
        # 											columns=self.classes)
        # 	# print(self.W_outputs_to_output)
        # 	# print(self.W_outputs_to_output.shape)
        # 	self.b_out = pd.DataFrame(np.zeros(len(self.classes)),
        # 								index=self.classes,
        # 								columns=["CLASS_BIASES"])
        # else:
        # 	self.W_outputs_to_output = pd.DataFrame(np.concatenate([np.eye(tree.C)
        # 															for tree in self.ndts],
        # 															axis=0),
        # 											index=[str(t.tree_id)
        # 												   for t in self.ndts],
        # 											columns=["MEAN VALUES"])
        # 	print(self.W_outputs_to_output)
        # 	# print(self.W_outputs_to_output.shape)
        # 	self.b_out = pd.DataFrame(np.zeros(1),
        # 							  columns=["BIAS"])

    def to_keras(self, loss,
                 metrics=[], optimizer=optimizers.Adam,
                 kernel_regularizer=[None, None, None],
                 optimizer_params={"lr": 0.001, "beta_1": 0.9,
                                   "beta_2": 0.999, "epsilon": 1e-8,
                                   "decay": 1e-6}):
        """
        Creates keras NN model
        """

        self.input_layer = Input(shape=(self.D,))
        self.count_ops = 0

        for tree in self.ndts:
            tree.to_keras(loss=loss, metrics=metrics,
                          optimizer=optimizer, kernel_regularizer=kernel_regularizer,
                          optimizer_params=optimizer_params)
            self.count_ops = self.count_ops+tree.count_ops

        # # Define the averaging model
        # if len(tree_models) > 1:
        # 	self.concatenation_output_layers = Concatenate()([output
        # 											for output in tree_models])
        # else:
        # 	self.concatenation_output_layers = tree_models[0]
        # self.dropouts_output_layers = Dropout(dropouts[2])(self.concatenation_output_layers)
        # if self.is_classifier:
        # 	self.output_layer = Dense(len(self.classes),
        # 							  activation='softmax')(self.dropouts_output_layers)
        # else:
        # 	self.output_layer = Dense(1)(self.dropouts_output_layers)

        # self.model = Model(inputs=self.input_layer, outputs=self.output_layer)
        # self.sgd = optimizer(**optimizer_params)
        # self.model.compile(loss=loss, optimizer=self.sgd, metrics=metrics)
        # # print(self.model.summary())
        # # plot_model(self.model, to_file='model.png')
        # # print(self.model.layers[2].layers[2])

        # self.count_ops = self.count_ops + \
        # 	self.model.layers[-1].input_shape[1]*self.model.layers[-1].output_shape[1]
        # if self.is_classifier:
        # 	self.count_ops = self.count_ops+self.model.layers[-1].output_shape[1] *\
        # 						(11+10*len(self.classes))
        # else:
        # 	self.count_ops = self.count_ops+self.model.layers[-1].output_shape[1]

        # self.model.layers[-1].set_weights(weights=[self.W_outputs_to_output,
        # 								  self.b_out.values.flatten()])

    def fit(self, X, y, sparse=False, epochs=100, min_delta=0, patience=10,
            earlyStopping=True, monitor="loss", validation_data=None,
            restore_best_weights=True, verbose=0, **fit_params):

        h = []

        for nn in self.ndts:
            h.append(nn.fit(X, y, sparse=sparse, epochs=epochs, min_delta=min_delta,
                     patience=patience, earlyStopping=earlyStopping, monitor=monitor,
                     validation_data=validation_data, restore_best_weights=restore_best_weights,
                     verbose=verbose, **fit_params))

        return h

    def predict(self, X, **kwargs):

        preds = np.zeros((X.shape[0], len(self.ndts)))
        for i, nn in enumerate(self.ndts):
            p = nn.predict(X, **kwargs)
            preds[:, i] = p.reshape((p.shape[0],))
        preds = np.mean(preds, axis=1)

        return preds

    def get_activations(self,
                        X,
                        y=None):
        """
        Get activation of each layer of the NRF

        Args:
                X (numpy.array or pandas.DataFrame): Training set
                y (numpy.array or pandas.Series): training set labels

        Returns:
                list: activations for each neural decision tree and for the last layer
                of the neural random forest
        """
        output_a = pd.DataFrame(self.model.predict(X),
                                columns=self.b_out.columns)
        return [t.get_activations(X, y) for t in self.ndts]+[[output_a]]

    def get_weights_from_NN(self):
        """
        Get the NRF weights from Keras object
        """
        for tree in self.ndts:
            tree.get_weights_from_NN()
        # print(len(self.model.layers))
        # w_9 = self.model.layers[-1].get_weights()
        # self.W_outputs_to_output_nn = pd.DataFrame(w_9[0],
        # 								index=[str(t.tree_id)+"_"+str(t_class)
        # 									   for t in self.ndts for t_class in t.classes],
        # 								columns=self.classes)
        # self.b_out_nn = pd.DataFrame(w_9[1],
        # 							 index=self.classes,
        # 							 columns=["CLASS_BIASES"])

    def compute_weights_differences(self):
        """
        Compute the difference between the weights mapped from the original RF,
        and the NRF weights

        Returns:
                list: list of pandas.DataFrames containing the weight differences
        """
        # TODO
        self.get_weights_from_NN()
        # for tree in self.ndts:
        # 	tree.compute_weights_differences()
        # self.diff_W_outputs_to_output_nn = self.W_outputs_to_output -\
        # 								   self.W_outputs_to_output_nn
        # self.diff_b_out_nn = self.b_out_nn - self.b_out

        return [t.compute_weights_differences() for t in self.ndts]  # +\
        #    [[diff_W_outputs_to_output_nn, diff_b_out_nn]]

    def print_tree_weights(self):
        """
        Print the weights mapped from the original RF
        """
        for t in self.ndts:
            t.print_tree_weights()
        # print("W: outputs -> output")
        # print(self.W_outputs_to_output)
        # print("b: output")
        # print(self.b_out)

    def print_nn_weights(self):
        """
        Print the NRF weights
        """
        for t in self.ndts:
            t.print_nn_weights()
        # print("W: outputs -> output")
        # print(self.W_outputs_to_output_nn)
        # print("b: output")
        # print(self.b_out_nn)


class NRFindClassifier(nrf_independent_ndt, NDTClassifier):
    def __init__(self, D, gammas=[100, 1], sigma=0, gamma_activation=True):
        super().__init__(D=D, gammas=gammas, sigma=sigma, gamma_activation=gamma_activation,
                         is_classifier=True)

    def predict_probas(self, X, **kwargs):
        return super().predict(X, **kwargs)


class NRFindRegressor(nrf_independent_ndt, NDTRegressor):
    def __init__(self, D, gammas=[100, 1], sigma=0, gamma_activation=True):
        super().__init__(D=D, gammas=gammas, sigma=sigma, gamma_activation=gamma_activation,
                         is_classifier=False)


#### nrf_sf_ind
class nrf_sf_ind(nrf_independent_ndt):

    def __init__(self, D, gammas=[10, 1]):
        nrf_independent_ndt.__init__(self, D, gammas)

    def compute_matrices_and_biases(self, forest):
        """
        Computes the weights and biases of the NRF_SF.
        Each SF from the Forest is mapped to a NRF, and a last layer is created for the weighted voting.

        Args:
        forest (Forest): Forest generated from an RF or a Bagging classifier
        """
        self.f = forest
        self.classes = self.f.classes_
        self.C = len(self.classes)
        self.nrfs = []
        for i, sf in enumerate(forest.subforests):
            self.nrfs.append(nrf_independent_ndt(self.D, self.gammas))
            self.nrfs[i].compute_matrices_and_biases(sf)
        # Define averaging layer
        self.W_outputs_to_output = pd.DataFrame(np.concatenate([np.eye(self.C)*self.gammas[-1] for subf in self.nrfs], axis=0),
                                                # index=[str(s.tree_id)+"_"+str(c) for s in self.nrfs for s_class in s.classes],
                                                columns=self.classes)
        self.b_out = pd.DataFrame(np.zeros(self.C),
                                    index=self.classes,
                                    columns=["CLASS_BIASES"])

    def to_keras(self, loss='categorical_crossentropy',
                 metrics=[], optimizer=optimizers.Adam, kernel_regularizer=[l1(0), l1(0), l1(0)],
                 optimizer_params={"lr": 0.001, "beta_1": 0.9,
                                   "beta_2": 0.999, "epsilon": 1e-8,
                                   "decay": 1e-6}):
        """
        Creates keras NN model
        """

        self.input_layer = Input(shape=(self.D,))
        self.count_ops = 0

        sf_models = []
        for sf in self.nrfs:
            sf.to_keras(dloss, metrics, optimizer, kernel_regularizer, optimizer_params)
            sf_models.append(sf.model(self.input_layer))
            self.count_ops = self.count_ops+sf.count_ops

        # Define the averaging model
        if len(sf_models) > 1:
            self.concatenation_output_layers = Concatenate()([output for output in sf_models])
        else:
            self.concatenation_output_layers = sf_models[0]

        self.output_layer = Dense(self.C, activation='softmax')(self.concatenation_output_layers)

        self.model = Model(inputs=self.input_layer, outputs=self.output_layer)

        plot_model(self.model)
        # print(self.model.summary())

        self.sgd = optimizer(**optimizer_params)
        self.model.compile(loss=loss, optimizer=self.sgd, metrics=metrics)

        self.count_ops = self.count_ops + \
            self.model.layers[-1].input_shape[1]*self.model.layers[-1].output_shape[1]
        self.count_ops = self.count_ops+self.model.layers[-1].output_shape[1]*(11+10*len(self.classes))

        self.model.layers[-1].set_weights(weights=[self.W_outputs_to_output, self.b_out.values.flatten()])

    def get_activations(self,
                        X,
                        y=None):
        """
        Get activation of each layer of the NRF

        Args:
                X (numpy.array or pandas.DataFrame): Training set
                y (numpy.array or pandas.Series): training set labels

        Returns:
                list: activations for each neural decision tree and for the last layer of the neural random forest
        """
        output_a = pd.DataFrame(self.model.predict(X), columns=self.b_out.columns)
        return [sf.get_activations(X, y) for sf in self.nrfs]+[[output_a]]

    def get_weights_from_NN(self):
        """
        Get the NRF weights from Keras object
        """
        for tree in self.ndts:
            tree.get_weights_from_NN()
        w_9 = self.model.layers[9].get_weights()
        self.W_outputs_to_output_nn = pd.DataFrame(w_9[0],
                                        index=[str(t.tree_id)+"_"+str(c) for t in self.ndts for t_class in t.classes],
                                        columns=self.classes)
        self.b_out_nn = pd.DataFrame(w_9[1],
                                    index=self.classes,
                                    columns=["CLASS_BIASES"])


class nrf_sf_full(nrf_independent_ndt):
    def __init__(self, D, gammas=[10, 1]):
        nrf_independent_ndt.__init__(self, D, gammas)

    def compute_matrices_and_biases(self, forest):

        self.f = forest
        self.classes = self.f.classes_
        self.nrfs = []
        for i, sf in enumerate(forest.subforests):
            self.nrfs.append(nrf_fully_connected(self.D, self.gammas, i))
            self.nrfs[i].compute_matrices_and_biases(sf)
        # Define averaging layer
        # self.W_outputs_to_output = pd.DataFrame(np.concatenate([np.eye(sf.C)*self.gammas[-1] for sf in self.nrfs], axis=0),
        # 										  index=[str(sf.tree_id)+"_"+str(sf_class) for sf in self.nrfs for sf_class in sf.classes],
        # 										  columns=self.classes)

        self.W_outputs_to_output = pd.DataFrame(np.concatenate([np.eye(sf.C) for sf in self.nrfs], axis=0),
                                                index=[str(sf.tree_id)+"_"+str(sf_class) for sf in self.nrfs for sf_class in sf.classes],
                                                columns=self.classes)

        self.b_out = pd.DataFrame(np.zeros(len(self.classes)),
                                    index=self.classes,
                                    columns=["CLASS_BIASES"])

    def to_keras(self, loss='categorical_crossentropy',
                metrics=[], optimizer=optimizers.Adam, kernel_regularizer=[l1(0), l1(0), l1(0)],
                optimizer_params={"lr": 0.001, "beta_1": 0.9,
                                  "beta_2": 0.999, "epsilon": 1e-8,
                                  "decay": 1e-6}):
        """
        Creates keras NN model
        """

        self.input_layer = Input(shape=(self.D,))
        self.count_ops = 0

        sf_models = []
        for sf in self.nrfs:
            sf.to_keras(loss, metrics, optimizer, kernel_regularizer, optimizer_params)
            sf_models.append(sf.model(self.input_layer))
            self.count_ops = self.count_ops+sf.count_ops

        # Define the averaging model

        if len(sf_models) > 1:
            self.concatenation_output_layers = Concatenate()([output for output in sf_models])
        else:
            self.concatenation_output_layers = sf_models[0]
        self.output_layer = Dense(len(self.classes), activation='softmax')(self.concatenation_output_layers)

        self.model = Model(inputs=self.input_layer, outputs=self.output_layer)
        self.sgd = optimizer(**optimizer_params)
        self.model.compile(loss=loss, optimizer=self.sgd, metrics=metrics)

        self.count_ops = self.count_ops + \
            self.model.layers[-1].input_shape[1]*self.model.layers[-1].output_shape[1]
        self.count_ops = self.count_ops+self.model.layers[-1].output_shape[1]*(11+10*len(self.classes))

        # print(self.model.summary())
        self.model.layers[-1].set_weights(weights=[self.W_outputs_to_output,
                                                   self.b_out.values.flatten()])

    def get_weights_from_NN(self):
        """
        Get the NRF weights from Keras object
        """
        for sf in self.nrfs:
            sf.get_weights_from_NN()
        w_9 = self.model.layers[-1].layers[-1].get_weights()
        self.W_outputs_to_output_nn = pd.DataFrame(w_9[0],
                                        index=[str(sf.tree_id)+"_"+str(t_class) for sf in self.nrfs for t_class in sf.classes],
                                        columns=self.classes)
        self.b_out_nn = pd.DataFrame(w_9[1],
                                       index=self.classes,
                                       columns=["CLASS_BIASES"])

    def compute_weights_differences(self):
        """
        Compute the difference between the weights mapped from the original RF, and the NRF weights

        Returns:
                list: list of pandas.DataFrames containing the weight differences
        """
        self.get_weights_from_NN()
        diff_W_outputs_to_output_nn = self.W_outputs_to_output - self.W_outputs_to_output_nn
        diff_b_out_nn = self.b_out_nn - self.b_out
        return [sf.compute_weights_differences() for sf in self.nrfs] + [[diff_W_outputs_to_output_nn, diff_b_out_nn]]
if __name__ == "__main__":
	np.random.seed(0)
	from keras.regularizers import l1  #, l2
	import matplotlib.pyplot as plt
	dataset_length = 1000
	D = 2
	X = np.random.randn(dataset_length, D)*0.1
	X[0:dataset_length//2, 0] += 0.1
	X[0:dataset_length//2, 0] += 0.2
	Y = np.ones(dataset_length)
	Y[0:dataset_length//2] *= 0

	X_test = np.random.randn(dataset_length, D)*0.1
	X_test[0:dataset_length//2, 0] += 0.1
	X_test[0:dataset_length//2, 0] += 0.2
	Y_test = np.ones(dataset_length)
	Y_test[0:dataset_length//2] *= 0
	# Train a Tree
	rf = RandomForestClassifier(max_depth=5, n_estimators=20)
	rf.fit(X, Y)

	f = Forest(rf)
	f.sampling(DTSamplingFeatureImp, n_trees=5, n_groups=4)

	# a = nrf_fully_connected(D=2, gammas=[10, 1], sigma=0.)
	# a = nrf_independent_ndt(D=2, gammas=[10, 1])
	a = nrf_sf_full(D=2, gammas=[0.01, 0.01])
	# a = nrf_sf_ind(D=2, gammas=[10, 1])
	b = nrf_fully_connected(D=2, gammas=[10, 1], sigma=0.)
	a.compute_matrices_and_biases(f)
	b.compute_matrices_and_biases(rf)
	l1_coef = -4
	a.to_keras(kernel_regularizer=[l1(10**l1_coef), l1(10**l1_coef), l1(10**l1_coef)],
						 dropouts=[0.1, 0.1, 0.1])
	b.to_keras(kernel_regularizer=[l1(10**l1_coef), l1(10**l1_coef), l1(10**l1_coef)],
						 dropouts=[0.1, 0.1, 0.1])
	print(a.count_ops)
	print(b.count_ops)

	print("scores before training")
	print("SF test:", a.score(X_test, Y_test))
	print("SF train:", a.score(X, Y))
	print("RF test:", b.score(X_test, Y_test))
	print("RF train:", b.score(X, Y))
	errors_a = a.fit(X, Y, epochs=100)["loss"]
	# print(errors_a)
	plt.figure()
	plt.plot(errors_a)
	plt.title("errors SF")
	errors_b = b.fit(X, Y, epochs=100)["loss"]
	plt.figure()
	plt.plot(errors_b)
	plt.title("errors RF")

	print("scores after training")
	print("SF test:", a.score(X_test, Y_test))
	print("SF train:", a.score(X, Y))
	print("RF test:", b.score(X_test, Y_test))
	print("RF train:", b.score(X, Y))

	print("scores forest")
	print("SF test:", a.f.score(X_test, Y_test))
	print("SF train:", a.f.score(X, Y))
	print("RF test:", b.rf.score(X_test, Y_test))
	print("RF train:", b.rf.score(X, Y))
	# a.get_weights_from_NN()

	# print "Tree weights"
	# a.print_tree_weights()
	# print "NN weights"
	# a.print_nn_weights()
	# print "activations"
	# print a.get_activations(X)
	# differences = a.compute_weights_differences()
	# print(differences)
	# a.plot_old_new_network()
	# a.plot_differences()
	# a.plot_W_nn_quantiles()
	plt.show()



import numpy as np
from sklearn.ensemble import RandomForestClassifier, RandomForestRegressor
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor


def generateNewDataClassification(oracle, X, y, ratio, replace_labels=False, by_class=True, **params):
	'''
	Generate new data using an oracle to label randomly generated examples.
	The quantity of new data depends on the input ratio
	'''

	newX = np.copy(X)
	classes = np.unique(y)

	if hasattr(oracle, "predict_classes"):
		predict_func = oracle.predict_classes
	else:
		predict_func = oracle.predict

	if replace_labels:
		newY = predict_func(X, **params)
		if not isinstance(newY, list):
			newY = np.reshape(newY, (newY.shape[0],))

	else:
		newY = np.copy(y)

	if ratio > 0:

		if by_class:  #### distinguish between classes when sampling values for new examples

			for c in classes:
				idx = np.nonzero(y == c)[0]
				X_cl = X[idx, :]

				size = np.ceil(ratio*X_cl.shape[0]).astype(int)  ### at least one example from each class

				extra_X = np.zeros((size, X_cl.shape[1]))

				idx = np.random.choice(X_cl.shape[0], size=extra_X.shape,
									   replace=True)

				extra_X = X_cl[idx, np.arange(X_cl.shape[1])[None, :]]

				newX = np.vstack((newX, extra_X))
				# print(predict_func)
				preds = predict_func(extra_X, **params)
				if not isinstance(preds, list):
					if len(preds.shape) > 1:
						if preds.shape[1] == 1:
							preds = preds.reshape((preds.shape[0],))
					# print(preds.shape)
					newY = np.concatenate((newY, preds))  ### generate labels for
														  ### the new examples
				else:
					newY = np.concatenate((newY, preds))

		else:

			size = np.ceil(ratio*X.shape[0]).astype(int)  ### at least one example from each class

			extra_X = np.zeros((size, X.shape[1]))

			idx = np.random.choice(X.shape[0], size=extra_X.shape, replace=True)

			extra_X = X[idx, np.arange(X.shape[1])[None, :]]

			newX = np.vstack((newX, extra_X))
			preds = predict_func(extra_X, **params)
			preds = np.reshape(preds, (preds.shape[0],))
			newY = np.concatenate((newY, preds))  ### generate labels for
												  ### the new examples

	classes_Ynew = np.unique(newY)
	not_represented = set(classes) - set(classes_Ynew)

	if bool(not_represented):
		for c in not_represented:
			idx = np.nonzero(y == c)[0]
			# rand_idx = np.random.choice(idx, size=1)

			newX = np.vstack((newX, X[idx, :]))
			newY = np.concatenate((newY, y[idx]))

	idx = np.arange(len(newY))
	np.random.shuffle(idx)  ### shuffle examples
	newX = newX[idx, :]
	newY = np.array([newY[i] for i in idx])

	# print("#################")
	# print(idx)
	# print(newX.shape)
	# print(len(newY))

	return(newX, newY)


def generateNewDataRegression(oracle, X, y, ratio, replace_labels=False, by_class=None, **params):
	'''
	Generate new data using an oracle to label randomly generated examples.
	The quantity of new data depends on the input ratio
	'''

	newX = np.copy(X)

	if replace_labels:
		newY = oracle.predict(X, **params)
		newY = np.reshape(newY, (newY.shape[0],))

	else:
		newY = np.copy(y)

	if ratio > 0:

		size = np.ceil(ratio*X.shape[0]).astype(int)  ### at least one example from each class

		extra_X = np.zeros((size, X.shape[1]))

		idx = np.random.choice(X.shape[0], size=extra_X.shape, replace=True)

		extra_X = X[idx, np.arange(X.shape[1])[None, :]]

		newX = np.vstack((newX, extra_X))

		preds = oracle.predict(extra_X, **params)
		preds = np.reshape(preds, (preds.shape[0],))
		newY = np.concatenate((newY, preds))  ### generate labels for
											  ### the new examples

	idx = np.arange(len(newY))
	np.random.shuffle(idx)  ### shuffle examples
	newX = newX[idx, :]
	newY = newY[idx]

	return(newX, newY)


def DT_extraction(oracle, X, Y, ratio, max_depth, n_trees=1,
				  replace_labels=False, by_class=True,
				  classification=True, **params):

	if hasattr(oracle, "_estimator_type"):
		if oracle._estimator_type == "classifier":
			generateNewData = generateNewDataClassification
			DecisionTree = DecisionTreeClassifier
			RandomForest = RandomForestClassifier
		else:
			generateNewData = generateNewDataRegression
			DecisionTree = DecisionTreeRegressor
			RandomForest = RandomForestRegressor

	else:
		if classification:
			generateNewData = generateNewDataClassification
			DecisionTree = DecisionTreeClassifier
			RandomForest = RandomForestClassifier
		else:
			generateNewData = generateNewDataRegression
			DecisionTree = DecisionTreeRegressor
			RandomForest = RandomForestRegressor

	newX, newY = generateNewData(oracle, X, Y, ratio, replace_labels, by_class, **params)

	if n_trees == 1:
		tree = DecisionTree(max_depth=max_depth, random_state=0)
		tree.fit(newX, newY)

		return tree

	else:
		trees = RandomForest(n_estimators=n_trees, max_depth=max_depth, random_state=0)
		trees.fit(newX, newY)

		return trees



from Neural_Decision_Tree import NDTClassifier
from sklearn.tree import DecisionTreeClassifier
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sys
from keras.regularizers import l1
sys.path.append("../data_mngmt")
from data import load_letter
from keras.utils import to_categorical

X_s, X_t_train, X_t_test, y_s, y_t_train, y_t_test, = load_letter()
X_t = np.concatenate([X_t_train, X_t_test])
y_t = np.concatenate([y_t_train, y_t_test])
y_t_cat = to_categorical(y_t)
y_s_cat = to_categorical(y_s)


results = []

for i in range(110, 210, 10):
	print(i)
	X_t_train = X_t[:i, :]
	X_t_test = X_t[i:, :]
	y_t_train = y_t[:i]
	y_t_test = y_t[i:]
	y_t_train_cat = y_t_cat[:i, :]
	y_t_test_cat = y_t_cat[i:, :]
	for depth in range(2,  5):
		print(depth)
		for rep in range(3):
			print(rep)
			# create tree
			clf = DecisionTreeClassifier(max_depth=depth)
			clf = clf.fit(X_s, y_s)
			# create NDT
			a = NDTClassifier(D=X_s.shape[1], gammas=[1, 1, 1], tree_id=0)
			a.compute_matrices_and_biases(clf)
			a.to_keras(dropouts=[0., 0., 0.], kernel_regularizer=[l1(1e-10), l1(1e-10), l1(1e-10)])
			# compute scores
			t_source_only_tree = clf.score(X_t_test, y_t_test)
			t_source_only_NN_before_training = a.score(X_t_test, y_t_test)
			s_tree = clf.score(X_s, y_s)
			s_NN_before_training = a.score(X_s, y_s)
			# Fit NN
			errors = a.fit(X_s, y_s_cat, epochs=100, to_categorical_conversion=False)
			# compute new scores
			t_source_only_NN_after_training = a.score(X_t_test, y_t_test)
			s_NN_after_training = a.score(X_s, y_s)
			# refine on target
			errors = a.fit(X_t_train, y_t_train_cat, epochs=100, to_categorical_conversion=False)
			clf_t = DecisionTreeClassifier(max_depth=depth)
			clf_t = clf_t.fit(X_t_train,  y_t_train)
			target_only_tree = clf_t.score(X_t_test, y_t_test)
			clf_sl = DecisionTreeClassifier(max_depth=depth)
			clf_sl = clf_sl.fit(X_t_test,  y_t_test)
			skyline = clf_sl.score(X_t_test,  y_t_test)
			transfered_NN = a.score(X_t_test, y_t_test)
			transfered_NN_source = a.score(X_s, y_s)
			current_result = [i, 
							depth, 
							s_tree, 
							s_NN_before_training, 
							s_NN_after_training,
							t_source_only_tree,
							t_source_only_NN_before_training,
							t_source_only_NN_after_training,
							target_only_tree,
							skyline,
							transfered_NN,
							transfered_NN_source
							]
			results.append(current_result)
			print(current_result)
res = pd.DataFrame(results, columns=["size",
										"depth",
										"s_tree",
										"s_NN_before_training",
										"s_NN_after_training",
										"t_source_only_tree",
										"t_source_only_NN_before_training",
										"t_source_only_NN_after_training",
										"target_only_tree",
										"skyline",
										"transfered_NN",
										"transfered_NN_source"])

res_d_2 = res[res["depth"] == 2]
res_d_3 = res[res["depth"] == 3]
res_d_4 = res[res["depth"] == 4]


plt.plot(res_d_2["size"], res_d_2["t_source_only_tree"], "yo--")
plt.plot(res_d_2["size"], res_d_2["t_source_only_NN_after_training"], "yo-")
plt.plot(res_d_2["size"], res_d_2["target_only_tree"], "ro--")
plt.plot(res_d_2["size"], res_d_2["skyline"], "ko-")
plt.plot(res_d_2["size"], res_d_2["transfered_NN"], "go-")
plt.show()


plt.plot(res_d_3["size"], res_d_3["t_source_only_tree"], "yo--")
plt.plot(res_d_3["size"], res_d_3["t_source_only_NN_after_training"], "yo-")
plt.plot(res_d_3["size"], res_d_3["target_only_tree"], "ro--")
plt.plot(res_d_3["size"], res_d_3["skyline"], "ko-")
plt.plot(res_d_3["size"], res_d_3["transfered_NN"], "go-")
plt.show()


plt.plot(res_d_4["size"], res_d_4["t_source_only_tree"], "yo--")
plt.plot(res_d_4["size"], res_d_4["t_source_only_NN_after_training"], "yo-")
plt.plot(res_d_4["size"], res_d_4["target_only_tree"], "ro--")
plt.plot(res_d_4["size"], res_d_4["skyline"], "ko-")
plt.plot(res_d_4["size"], res_d_4["transfered_NN"], "go-")
plt.show()



"""
Discretizers classes, to be used in lime_tabular
"""
import numpy as np
import sklearn
import sklearn.tree
import scipy
from sklearn.utils import check_random_state
from abc import ABCMeta, abstractmethod


class BaseDiscretizer():
    """
    Abstract class - Build a class that inherits from this class to implement
    a custom discretizer.
    Method bins() is to be redefined in the child class, as it is the actual
    custom part of the discretizer.
    """

    __metaclass__ = ABCMeta  # abstract class

    def __init__(self, data, categorical_features, feature_names, labels=None, random_state=None,
                 data_stats=None):
        """Initializer
        Args:
            data: numpy 2d array
            categorical_features: list of indices (ints) corresponding to the
                categorical columns. These features will not be discretized.
                Everything else will be considered continuous, and will be
                discretized.
            categorical_names: map from int to list of names, where
                categorical_names[x][y] represents the name of the yth value of
                column x.
            feature_names: list of names (strings) corresponding to the columns
                in the training data.
            data_stats: must have 'means', 'stds', 'mins' and 'maxs', use this
                if you don't want these values to be computed from data
        """
        self.to_discretize = ([x for x in range(data.shape[1])
                               if x not in categorical_features])
        self.data_stats = data_stats
        self.names = {}
        self.lambdas = {}
        self.means = {}
        self.stds = {}
        self.mins = {}
        self.maxs = {}
        self.random_state = check_random_state(random_state)

        # To override when implementing a custom binning
        bins = self.bins(data, labels)
        bins = [np.unique(x) for x in bins]

        # Read the stats from data_stats if exists
        if data_stats:
            self.means = self.data_stats.get("means")
            self.stds = self.data_stats.get("stds")
            self.mins = self.data_stats.get("mins")
            self.maxs = self.data_stats.get("maxs")

        for feature, qts in zip(self.to_discretize, bins):
            n_bins = qts.shape[0]  # Actually number of borders (= #bins-1)
            boundaries = np.min(data[:, feature]), np.max(data[:, feature])
            name = feature_names[feature]

            self.names[feature] = ['%s <= %.2f' % (name, qts[0])]
            for i in range(n_bins - 1):
                self.names[feature].append('%.2f < %s <= %.2f' %
                                           (qts[i], name, qts[i + 1]))
            self.names[feature].append('%s > %.2f' % (name, qts[n_bins - 1]))

            self.lambdas[feature] = lambda x, qts=qts: np.searchsorted(qts, x)
            discretized = self.lambdas[feature](data[:, feature])

            # If data stats are provided no need to compute the below set of details
            if data_stats:
                continue

            self.means[feature] = []
            self.stds[feature] = []
            for x in range(n_bins + 1):
                selection = data[discretized == x, feature]
                mean = 0 if len(selection) == 0 else np.mean(selection)
                self.means[feature].append(mean)
                std = 0 if len(selection) == 0 else np.std(selection)
                std += 0.00000000001
                self.stds[feature].append(std)
            self.mins[feature] = [boundaries[0]] + qts.tolist()
            self.maxs[feature] = qts.tolist() + [boundaries[1]]

    @abstractmethod
    def bins(self, data, labels):
        """
        To be overridden
        Returns for each feature to discretize the boundaries
        that form each bin of the discretizer
        """
        raise NotImplementedError("Must override bins() method")

    def discretize(self, data):
        """Discretizes the data.
        Args:
            data: numpy 2d or 1d array
        Returns:
            numpy array of same dimension, discretized.
        """
        ret = data.copy()
        for feature in self.lambdas:
            if len(data.shape) == 1:
                ret[feature] = int(self.lambdas[feature](ret[feature]))
            else:
                ret[:, feature] = self.lambdas[feature](
                    ret[:, feature]).astype(int)
        return ret

    def get_undiscretize_values(self, feature, values):
        mins = np.array(self.mins[feature])[values]
        maxs = np.array(self.maxs[feature])[values]

        means = np.array(self.means[feature])[values]
        stds = np.array(self.stds[feature])[values]
        minz = (mins - means) / stds
        maxz = (maxs - means) / stds
        min_max_unequal = (minz != maxz)

        ret = minz
        ret[np.where(min_max_unequal)] = scipy.stats.truncnorm.rvs(
            minz[min_max_unequal],
            maxz[min_max_unequal],
            loc=means[min_max_unequal],
            scale=stds[min_max_unequal],
            random_state=self.random_state
        )
        return ret

    def undiscretize(self, data):
        ret = data.copy()
        for feature in self.means:
            if len(data.shape) == 1:
                ret[feature] = self.get_undiscretize_values(
                    feature, ret[feature].astype(int).reshape(-1, 1)
                )
            else:
                ret[:, feature] = self.get_undiscretize_values(
                    feature, ret[:, feature].astype(int)
                )
        return ret


class StatsDiscretizer(BaseDiscretizer):
    """
        Class to be used to supply the data stats info when discretize_continuous is true
    """

    def __init__(self, data, categorical_features, feature_names, labels=None, random_state=None,
                 data_stats=None):

        BaseDiscretizer.__init__(self, data, categorical_features,
                                 feature_names, labels=labels,
                                 random_state=random_state,
                                 data_stats=data_stats)

    def bins(self, data, labels):
        bins_from_stats = self.data_stats.get("bins")
        bins = []
        if bins_from_stats is not None:
            for feature in self.to_discretize:
                bins_from_stats_feature = bins_from_stats.get(feature)
                if bins_from_stats_feature is not None:
                    qts = np.array(bins_from_stats_feature)
                    bins.append(qts)
        return bins


class QuartileDiscretizer(BaseDiscretizer):
    def __init__(self, data, categorical_features, feature_names, labels=None, random_state=None):

        BaseDiscretizer.__init__(self, data, categorical_features,
                                 feature_names, labels=labels,
                                 random_state=random_state)

    def bins(self, data, labels):
        bins = []
        for feature in self.to_discretize:
            qts = np.array(np.percentile(data[:, feature], [25, 50, 75]))
            bins.append(qts)
        return bins


class DecileDiscretizer(BaseDiscretizer):
    def __init__(self, data, categorical_features, feature_names, labels=None, random_state=None):
        BaseDiscretizer.__init__(self, data, categorical_features,
                                 feature_names, labels=labels,
                                 random_state=random_state)

    def bins(self, data, labels):
        bins = []
        for feature in self.to_discretize:
            qts = np.array(np.percentile(data[:, feature],
                                         [10, 20, 30, 40, 50, 60, 70, 80, 90]))
            bins.append(qts)
        return bins


class EntropyDiscretizer(BaseDiscretizer):
    def __init__(self, data, categorical_features, feature_names, labels=None, random_state=None):
        if(labels is None):
            raise ValueError('Labels must be not None when using \
                             EntropyDiscretizer')
        BaseDiscretizer.__init__(self, data, categorical_features,
                                 feature_names, labels=labels,
                                 random_state=random_state)

    def bins(self, data, labels):
        bins = []
        for feature in self.to_discretize:
            # Entropy splitting / at most 8 bins so max_depth=3
            dt = sklearn.tree.DecisionTreeClassifier(criterion='entropy',
                                                     max_depth=3,
                                                     random_state=self.random_state)
            x = np.reshape(data[:, feature], (-1, 1))
            dt.fit(x, labels)
            qts = dt.tree_.threshold[np.where(dt.tree_.children_left > -1)]

            if qts.shape[0] == 0:
                qts = np.array([np.median(data[:, feature])])
            else:
                qts = np.sort(qts)

            bins.append(qts)

        return bins



"""
Subforest class
"""

# import sys
from itertools import combinations

import numpy as np

# sys.path.append("../utils")

from .common import *
from ..utils.comparison_functions import fleiss_score
from ..DT_extraction import generateNewDataClassification

########## Subforests and Subforest ensembles


class Subforest:
	'''
	Methods for subforests (groups of DTS)
	'''

	def __init__(self, tree_list, classes):
		self.estimators_ = tree_list
		self.n_estimators = len(tree_list)
		self.n_features = tree_list[0].n_features_
		self.classes_ = classes

	def predict_proba(self, X):
		return predict_trees_proba(self.estimators_, X, len(self.classes_))

	def predict(self, X):
		return predict_trees(self.estimators_, X, self.classes_)

	def score(self, X, y):
		return score_trees(self.estimators_, X, y)

	def fleiss(self, X):
		return fleiss_score(self.estimators_, X)

	def feature_importances(self):
		importances = np.zeros((self.n_estimators, self.n_features))

		for i, t in enumerate(self.estimators_):
			importances[i, :] = t.feature_importances_

		return np.mean(importances, axis=0)

	def data_augmentation(self, X, y, ratio, replace_labels, by_class):
		return generateNewDataClassification(self, X, y, ratio, replace_labels, by_class)

	def newTree(self, X, y, ratio, replace_labels, by_class=True, random_state=0):
		'''
		Generate a new tree trained with extra data generated using the subforest
		as an oracle
		'''
		newX, newY = generateNewDataClassification(self, X, y, ratio, replace_labels, by_class)

		# feat_imp = self.feature_importances()
		# feat_imp = np.floor(feat_imp/sum(feat_imp)*100).astype(int)

		# newX = scale(newX)
		# print(feat_imp)
		# newX = newX*feat_imp
		# print("newY :", np.unique(newY))
		# tmp = np.zeros((newX.shape[0], sum(feat_imp)))

		# start = 0
		# end = 0
		# print(feat_imp)
		# for i, fi in enumerate(feat_imp):
		# 	end = start+fi
		# 	tmp[:, start:end] = np.tile(newX[:,i], (fi,1)).T
		# 	start = end
		# newX = tmp

		tree = DecisionTreeClassifier(random_state=random_state)
		tree.fit(newX, newY)

		return tree  #, feat_imp


def SFSamplingCombinations(groups, clusters, combination_length, n_trees, n_sf):
	'''
	Create comb(n_clusters, combination_length) subforests with
	n_trees each by radomly selecting trees from a combination of
	clusters. Size of the clusters is taken into account in the
	sampling process.
	'''

	combs = list(combinations(range(len(groups)), combination_length))
	idx = np.random.choice(len(combs), size=n_sf, replace=False)
	combs = [combs[i] for i in idx]

	subforests = []

	for c in combs:
		probas = [len(groups[k].estimators_) for k in c]  ### create the probability
		probas = [p/sum(probas) for p in probas]    ## distribution using the
												    ## sizes of the clusters

		r = np.random.choice(c, size=n_trees, p=probas)  ### choose n_trees
		subforests.append(Subforest([np.random.choice(groups[k].estimators_) for k in r], groups[0].classes_))

	return n_sf, subforests


def SFSamplingDistance(groups, clusters, distance_matrix, n_trees, n_sf):

	def single_linkage(clusters, distance_matrix):
		labs = set(clusters)
		cluster_dists = np.zeros((len(labs), len(labs)))

		for i, c1 in enumerate(labs):
			for j, c2 in enumerate(labs):
				if i != j:
					D = distance_matrix[clusters == c1, :]
					D = D[:, clusters == c2]

					cluster_dists[i, j] = np.min(D)
					cluster_dists[j, i] = cluster_dists[i, j]

		return cluster_dists

	labs = set(clusters)
	cluster_dists = single_linkage(clusters, distance_matrix)

	subforests = []

	select_groups = np.random.choice(len(groups), size=n_sf)

	for i in select_groups:
		d = 1-cluster_dists[i, :]
		probas = d/sum(d)
		# print(1-d)
		# print(probas)
		trees_cluster = np.random.choice(len(labs), size=n_trees, p=probas)

		trees_cluster, counts = np.unique(trees_cluster, return_counts=True)
		# print(trees_cluster)
		# print(counts)

		trees = []
		for j, cl in enumerate(trees_cluster):
			trees.extend(np.random.choice(groups[cl].estimators_, size=counts[j]))

		subforests.append(Subforest(trees, groups[i].classes_))

	return n_sf, subforests



# import sys

import numpy as np
from sklearn.metrics import accuracy_score

# sys.path.append("../../utils")

from .common import *
from .DT_grouping import *
from .Subforest import *
# from comparison_functions import fleiss_score
from ..utils.comparison_functions import fleiss_score


class Forest:
	'''
	Ensemble of subforests from a random forest
	'''

	def __init__(self, cl, n_trees=None, X=None):
		trees = cl.estimators_

		if n_trees is not None:
			if n_trees < len(cl.estimators_):
				
				trees_ = []
				for t in cl.estimators_:
					if len(np.unique(t.predict(X))) == t.n_classes_:
						trees_.append(t)
				
				if len(trees_) >= n_trees:
					trees = np.random.choice(trees_, size=n_trees, replace=False)
			
		self.estimators_ = trees
		self.n_estimators = len(trees)
		self.classes_ = cl.classes_
		self.n_features = cl.n_features_
		self.n_subforests = 1
		self.subforests = [Subforest(self.estimators_, self.classes_)]
		self.weights = []
		self.clusters = []
		self.representatives = None
	
	def clustering(self, function, **params):
		self.n_subforests, subforests, self.clusters,\
			self.weights, representatives = function(self.estimators_, **params)

		self.subforests = []
		for sf in subforests:
			self.subforests.append(Subforest(sf, self.classes_))

		self.representatives = Subforest(representatives, self.classes_)
	
	def sampling(self, function, **params):
		self.n_subforests, subforests = function(self.estimators_, **params)

		self.subforests = []
		for sf in subforests:
			self.subforests.append(Subforest(sf, self.classes_))
	
	def subforestSampling(self, function, **params):
		self.weights = []
		self.n_subforests, self.subforests = function(self.subforests, self.clusters, **params)

	def predict(self, X, subforests=False):
		'''
		Weighted vote using the subforests as learners
		'''

		prediction = [] 
		subforest_prediction = np.zeros((self.n_subforests, X.shape[0]))
		proba = np.zeros((X.shape[0], len(self.classes_)))

		if np.asarray(self.weights).size == 0:
			
			for i, sf in enumerate(self.subforests):
				subforest_prediction[i, :] = sf.predict(X)
				proba = proba+sf.predict_proba(X)
			
			proba = proba/self.n_subforests
		
		else:

			for i, sf in enumerate(self.subforests):
				subforest_prediction[i, :] = sf.predict(X)
				proba = proba+sf.predict_proba(X)*self.weights[i]
			
			proba = proba/sum(self.weights)

		prediction = np.take(self.classes_, np.argmax(proba, axis=1))

		if subforests:
			return prediction, subforest_prediction
		else:
			return prediction

	def score(self, X, y, subforests=False):
		'''
		Accuracy using subforests as learners
		'''
		
		if subforests:
			y_p, y_sf = self.predict(X, subforests)

			sf_accuracy = []
			for pred in y_sf:
				sf_accuracy.append(accuracy_score(y, pred))
			
			return accuracy_score(y, y_p), sf_accuracy
		
		else:
			y_p = self.predict(X, subforests)
			return accuracy_score(y, y_p)

	def fleiss(self, X):
		'''
		Fleiss score for each subforest
		'''

		f = []

		for sf in self.subforests:
			f.append(sf.fleiss(X))

		return f

	def mean_fleiss(self, X):
		if np.asarray(self.weights).size != 0:
			return np.average(self.fleiss(X), weights=self.weights)
		else:
			return np.average(self.fleiss(X), weights=None)
	
	def sf_fleiss(self, X):
		return fleiss_score(self.subforests, X)

	def feature_importances(self):
		importances = np.zeros((self.n_subforests, self.n_features))

		for i, sf in enumerate(self.subforests):
			importances[i, :] = sf.feature_importances()
		
		return importances, np.mean(importances, axis=0)

	def newTrees(self, X, y, ratio, replace_labels, by_class=True, random_state=0):
		tree_list = []

		for sf in self.subforests:
			t = sf.newTree(X, y, ratio, replace_labels, by_class, random_state=random_state)
			tree_list.append(t)
		
		return(Subforest(tree_list, self.classes_))

	def score_representatives(self, X, y):
		
		if self.representatives is not None:
			return self.representatives.score(X, y)




import numpy as np
from sklearn.metrics import accuracy_score

######### Functions for DT lists


def predict_trees_proba(tree_list, X, n_classes, weights=[]):
	'''
	Weighted probability predictions for DT lists (classification)
	'''

	probas = np.zeros((X.shape[0], n_classes))

	if np.asarray(weights).size == 0:

		for t in tree_list:
			# print(t.classes_)
			probas = probas+t.predict_proba(X)

		return probas/len(tree_list)

	else:
		for i, t in enumerate(tree_list):
			probas = probas+t.predict_proba(X)*weights[i]

		return probas/sum(weights)


def predict_trees(tree_list, X, classes, weights=[]):
	'''
	Weighted predictions for DT lists (classification)
	'''

	proba = predict_trees_proba(tree_list, X, len(classes), weights)

	prediction = np.take(classes, np.argmax(proba, axis=1))

	return prediction


def score_trees(tree_list, X, y, weights=[]):
	'''
	Weighted accuracy for a group of DT
	'''

	classes = np.unique(y)
	
	y_p = predict_trees(tree_list, X, classes, weights)
		
	return accuracy_score(y, y_p)




from itertools import combinations

import matplotlib.pyplot as plt
import numpy as np
from kmodes.kmodes import KModes
from matplotlib import cm
from scipy.cluster.hierarchy import dendrogram, fcluster, linkage
from scipy.spatial.distance import squareform
from sklearn.cluster import DBSCAN, AffinityPropagation, SpectralClustering

from .common import *

######### Forming groups of DTs


def avg_distance_to_clusters(tree, clusters, distance_matrix, n_clusters):
		'''
		Average distance from a tree to other clusters
		'''

		avg = 0

		for k in range(1, n_clusters+1):
			cluster_k = np.nonzero(clusters == k)[0]

			if tree not in cluster_k:
				distances = distance_matrix[tree, cluster_k]
				avg = avg+max(distances)
			
		if n_clusters > 1:					
			return avg/(n_clusters-1)
		else:
			return avg


def DTHAC(trees, distance_matrix, link_method="complete", n_clusters=None, X=None, y=None, 
		  plot=True, plot_dendrogram=False):
	'''
	Hierarchical Agglomerative Clustering for DTs
	as described in Design of Effective Multiple Classifier Systems by 
	Clustering of Classifiers; Giorgio Giacinto, Fabio Roli, and Giorgio Fumera
	'''

	def select_trees(distance_matrix, clusters, n_clusters):
		'''
		Selection of the trees that will not be pruned.
		From eah cluster, the tree with the maximum average distance to the other 
		clusters is selected.
		'''

		selected_trees = []
		for i in range(1, n_clusters+1):
			trees_in_cluster = np.nonzero(clusters == i)[0]
			max_dist = 0
			keep = -1
			
			for t in trees_in_cluster:
				d = avg_distance_to_clusters(t, clusters, distance_matrix, n_clusters)
				
				if d > max_dist:
					max_dist = d
					keep = t
	
			selected_trees.append(keep)
				
		return selected_trees
	
	dist_matrix = squareform(distance_matrix)
	link = linkage(dist_matrix, method=link_method)   ### calculate HAC tree

	if plot_dendrogram:
		plt.figure()
		dendrogram(link)
		
	if n_clusters:

		clusters = fcluster(link, n_clusters, criterion='maxclust')  ### form a maximum of n_clusters clusters
																	 ### can be inferior to n_clusters
		labs, weights = np.unique(clusters, return_counts=True)

		n_clusters_ = len(labs)
		representatives_idx = select_trees(distance_matrix, clusters, n_clusters_)
		representatives = [trees[i] for i in representatives_idx]
		
		clusters = clusters-1

		clusters_ = []
		for _ in range(n_clusters_):
			clusters_.append([])
		for i, c in enumerate(clusters):
			clusters_[c].append(trees[i])

		return n_clusters_, clusters_, clusters, weights, representatives

	else:   # choose the "best" nb clusters

		best_k = 0
		best_accuracy = -1
		accuracies = []
		nb_clusters = []

		for k in range(2, len(rf.estimators_)+1):
			# print("#########k:", k)
				
			clusters = fcluster(link, k, criterion='maxclust')
			labs, weights = np.unique(clusters, return_counts=True)
			w = weights

			if len(labs) == k:
				nb_clusters.append(k)
				selected_trees = select_trees(distance_matrix, clusters, len(labs))
				
				accuracy = score_trees([trees[j] for j in selected_trees], X, y, weights=w)
				accuracies.append(accuracy)

				if accuracy > best_accuracy:
					best_k = k
					best_accuracy = accuracy
					representatives_idx = selected_trees
					best_clusters = clusters
					best_weights = w		
		
		best_clusters = best_clusters-1
		
		clusters_ = [] 
		for _ in range(best_k):
			clusters_.append([])
		for i, c in enumerate(best_clusters):
			clusters_[c].append(trees[i])

		representatives = [trees[i] for i in representatives_idx]

		if plot:
			plt.plot(nb_clusters, accuracies, ".-", label="Clusters")
			plt.plot(best_k, best_accuracy, "*", label=("k = "+str(best_k) +
					"  acc = "+str(np.round(best_accuracy, decimals=2))))
			plt.ylim((0, 1.1))
			plt.xticks(nb_clusters[::2], nb_clusters[::2], rotation=45)
			plt.ylabel("Accuracy on validation set")
			plt.xlabel("Number of clusters")

		return best_k, clusters_, best_clusters, best_weights, representatives


def DTKmedoids(trees, distance_matrix, max_iter=100, n_clusters=None, X=None, y=None,  plot=True):
	'''
	K-medoids (Voronoi) algorithm for DTs
	Based on the implementation by Timo Erkkilä and Antti Lehmussola in
	scikit-learn's github repository
	'''

	def main_alg(n_clust, distance_matrix, max_iter):
		n_iter = 0
		medoids = np.random.permutation(distance_matrix.shape[0])[:n_clust]
		old_medoids = np.zeros((n_clust,))

		while not np.all(old_medoids == medoids) and n_iter < max_iter:
			n_iter = n_iter+1

			old_medoids = np.copy(medoids)
			clusters = np.argmin(distance_matrix[medoids, :], axis=0)

			for i in range(n_clust):
				current_cost = np.sum(distance_matrix[medoids[i], clusters == i])
					
				D = distance_matrix[clusters == i, :]
				D = D[:, clusters == i]

				all_costs = np.sum(D, axis=1)
				min_cost_idx = np.argmin(all_costs)
				min_cost = all_costs[min_cost_idx]

				if min_cost < current_cost:
					medoids[i] = np.where(clusters == i)[0][min_cost_idx]
			
		return medoids, clusters

	if n_clusters:

		representatives_idx, clusters = main_alg(n_clusters, distance_matrix, max_iter)
		_, weights = np.unique(clusters, return_counts=True)

		# clusters = clusters+1

		clusters_ = [] 
		for _ in range(n_clusters):
			clusters_.append([])
		for i, c in enumerate(clusters):
			clusters_[c].append(trees[i])

		representatives = [trees[i] for i in representatives_idx]
								
		return n_clusters, clusters_, clusters,  weights, representatives

	else:

		best_k = 0
		best_accuracy = -1
		# not_pruned = []
		best_clusters = []
		accuracies = []

		# distance_matrix = compare_trees_from_forest(rf, distance_function, **params)

		for k in range(2, len(trees)+1):
			# print("#########k:", k)

			medoids, clusters = main_alg(k, distance_matrix, max_iter)
				
			_, weights = np.unique(clusters, return_counts=True)
			
			accuracy = score_trees([trees[m] for m in medoids], X, y, weights=weights)
			accuracies.append(accuracy)

			if accuracy > best_accuracy:
				best_k = k
				best_accuracy = accuracy
				representatives_idx = medoids
				best_clusters = clusters
				best_weights = weights
		
		# best_clusters = best_clusters+1
		
		clusters_ = [] 
		for _ in range(best_k):
			clusters_.append([])
		for i, c in enumerate(best_clusters):
			clusters_[c].append(trees[i])

		representatives = [trees[i] for i in representatives_idx]

		if plot:
			plt.plot(range(2, len(trees)+1), accuracies, ".-", label="Subforests")
			plt.plot(best_k, best_accuracy, "*", 
					label=("k = "+str(best_k)+"  acc = "+str(np.round(best_accuracy, decimals=2))))
			plt.ylim((0, 1.1))
			plt.xticks(range(2, len(trees)+1)[::2], 
					   range(2, len(rf.estimators_)+1)[::2], rotation=45)	
			plt.ylabel("Accuracy on validation set")
			plt.xlabel("Number of clusters")

		return best_k, clusters_, best_clusters, best_weights, representatives


def DTDBSCAN(trees, distance_matrix, **params):

	db = DBSCAN(metric="precomputed", **params)
	db.fit(distance_matrix)

	clusters = db.labels_
	# print(db.labels_)

	n_clusters = len(set(clusters))  #- (1 if -1 in clusters else 0)
	n_noise_ = len(np.nonzero(clusters == -1)[0])
	print("n_clusters: ", n_clusters)
	print("%_noise: ", n_noise_/len(trees)*100)

	if n_noise_ > 0:
		clusters = clusters+1
	# else:
	# 	clusters = clusters+1
	_, weights = np.unique(clusters, return_counts=True)

	clusters_ = [] 
	for _ in range(n_clusters):
		clusters_.append([])
	for i, c in enumerate(clusters):
		clusters_[c].append(trees[i])

	if len(db.core_sample_indices_) == 0:
		representatives = [trees[0]]
	else:
		representatives = []
		for i in db.core_sample_indices_:
			representatives.append(trees[i])

	return n_clusters, clusters_, clusters,  weights, representatives


def DTAffinityPropagation(trees, distance_matrix, **params):
	affinity_matrix = 1-distance_matrix

	af = AffinityPropagation(affinity="precomputed", **params)
	af.fit(affinity_matrix)

	clusters = af.labels_
	# print(db.labels_)

	n_clusters = len(set(clusters))  #- (1 if -1 in clusters else 0)
	print("n_clusters: ", n_clusters)
	_, weights = np.unique(clusters, return_counts=True)

	# clusters = clusters+1

	clusters_ = [] 
	for _ in range(n_clusters):
		clusters_.append([])
	for i, c in enumerate(clusters):
		clusters_[c].append(trees[i])

	representatives = []
	for i in af.cluster_centers_indices_:
		representatives.append(trees[i])

	return n_clusters, clusters_, clusters,  weights, representatives


def DTSpectralClustering(trees, distance_matrix, **params):
	affinity_matrix = 1-distance_matrix

	sc = SpectralClustering(affinity="precomputed", **params)
	sc.fit(affinity_matrix)

	clusters = sc.labels_
	# print(db.labels_)

	n_clusters = len(set(clusters))  #- (1 if -1 in clusters else 0)
	print("n_clusters: ", n_clusters)
	_, weights = np.unique(clusters, return_counts=True)

	# clusters = clusters+1

	clusters_ = [] 
	for _ in range(n_clusters):
		clusters_.append([])
	for i, c in enumerate(clusters):
		clusters_[c].append(trees[i])

	representatives = [trees[0]]  # for compatibility

	return n_clusters, clusters_, clusters,  weights, representatives


def DTSamplingFeatureImp(trees, n_trees, n_groups):
	'''
	Generate groups of trees by sampling. Each group corresponds to a feature, and the
	trees inside each group are chosen randomly using their feature importances as probabilities
	'''

	groups = []

	distributions = np.zeros((len(trees), trees[0].n_features_))
	group_dists = [np.random.randint(trees[0].n_features_) for i in range(n_groups)]
	
	### each group corresponds to the distribution of one feature

	for i, t in enumerate(trees):
		distributions[i, :] = t.feature_importances_
		
	distributions = distributions/np.sum(distributions, axis=0)  ### feature importances to probability 
																 ### distributions

	for i in range(n_groups):
		idx = np.random.choice(len(trees), size=n_trees, 
			                   p=distributions[:, group_dists[i]])  ### choose  n_trees according to the 
							   									   ### distribution of the group
		groups.append([trees[k] for k in idx])
		
	return n_groups, groups
	
	
def DTSamplingUnif(trees, n_trees, n_groups):
	'''
	Generate groups of trees by sampling. Each group randomly chooses trees 
	with a uniform distribution.
	'''
	groups = []

	for _ in range(n_groups):
		idx = np.random.choice(len(trees), size=n_trees)  ### choose n_trees randomly
		groups.append([trees[k] for k in idx])
	
	return n_groups, groups


def DTCombinations(trees, combination_length):
	'''
	Generate groups of trees by combinations
	'''

	if combination_length == 0 or combination_length == len(trees):
		return 1, [trees]

	groups = []
	combs = list(combinations(trees, combination_length))

	for c in combs:
		groups.append(list(c))
	
	n_groups = len(groups)

	return n_groups, groups


########### Pruning


def keep_k_trees(tree_list, k, X, y):
	'''
	Order the trees according to their accuracies on X and keep the k
	most accurate.
	'''

	acc = []

	for t in tree_list:
		acc.append(t.score(X, y))
	
	idx = np.argsort(acc)

	keep = [tree_list[i] for i in idx[-k:]]

	return 1, [keep]


def CLUBDRF(tree_list, n_clusters, Xtr, Xval, yval):
	'''
	Implementation of the CLUBDRF algorithm introduced in 
	On Extreme Pruning of Random Forest Ensembles
	for Real-time Predictive Applications by
	Khaled Fawagreh, Mohamed Medhat Gaber, and Eyad Elyan, Member, IEEE
	'''

	def diversity(labels1, labels2, **_):  ## ratio of matching decisions to total decisions
		return np.sum(labels1 != labels2, axis=1)/labels1.shape[1]
	
	class_labels = np.zeros((len(tree_list), Xtr.shape[0]))
	for i, t in enumerate(tree_list):
		class_labels[i, :] = t.predict(Xtr)
	
	km = KModes(n_clusters=n_clusters, cat_dissim=diversity, n_init=1)  ### clustering by k-modes
	km.fit(class_labels)

	weights = np.unique(km.labels_, return_counts=True)[1]

	keep = []

	for c in range(n_clusters):  #### the representatives are the most accurate trees from each cluster
		best_score = 0
		best_tree = None
		
		idx = np.nonzero(km.labels_ == c)[0]
		trees = [tree_list[i] for i in idx]

		for t in trees:
			if t.score(Xval, yval) > best_score:
				best_tree = t
		
		keep.append(best_tree)

	# print(km.labels_)
	clusters = [] 
	for _ in range(n_clusters):
		clusters.append([])
	for i, c in enumerate(km.labels_):
		clusters[c].append(tree_list[i])

	return n_clusters, clusters, km.labels_, weights, keep

########## Clustering evaluation


def silhouette(clusters, distance_matrix, plot=True):
	silhouettes = silhouette_samples(distance_matrix, clusters, metric="precomputed")
	score = np.mean(silhouettes)

	if plot:
		plt.figure()

		n_clusters = len(set(clusters))

		y_lower = 10

		for i in range(n_clusters):
			# Aggregate the silhouette scores for samples belonging to
			# cluster i, and sort thems
			ith_cluster_silhouette_values = silhouettes[clusters == i]

			ith_cluster_silhouette_values.sort()

			size_cluster_i = ith_cluster_silhouette_values.shape[0]
			y_upper = y_lower + size_cluster_i

			color = cm.nipy_spectral(float(i) / n_clusters)
			plt.fill_betweenx(np.arange(y_lower, y_upper),
								0, ith_cluster_silhouette_values,
								facecolor=color, edgecolor=color, alpha=0.7)

			# Label the silhouette plots with their cluster numbers at the middle
			# plt.text(-0.075, y_lower + 0.5 * size_cluster_i, classes[i])

			# Compute the new y_lower for next plot
			y_lower = y_upper + 10  # 10 for the 0 samples

		plt.axvline(x=score, color="red", linestyle="--", label="Mean silhouette")
		plt.axvline(x=0, color="black", linestyle="-")

		plt.yticks([])  # Clear the yaxis labels / ticks
		plt.xticks([-0.1, 0, 0.2, 0.4, 0.6, 0.8, 1])
		plt.xlabel("Silhouette")
		plt.suptitle("Sample silhouettes")
		plt.legend()

	return score


def averageDistanceRatio(distance_matrix, clusters):
	d = 0
	l, counts = np.unique(clusters, return_counts=True)
	n_clusters = len(l)
	
	idx = np.triu_indices_from(distance_matrix, k=1)
	upper = distance_matrix[idx]
	d_tot = upper[upper.nonzero()].mean()
		
	for c in range(n_clusters):
		if counts[c] > 1:
			D = distance_matrix[clusters == c, :]
			D = D[:, clusters == c]
			
			idx = np.triu_indices_from(D, k=1)
			upper = D[idx]
			d = d+upper[upper.nonzero()].mean()
	
	d = d/n_clusters
	
	return d/d_tot



from setuptools import setup, find_packages

setup(
    name="lime_ndt",
    version="0.1.0",
    author="Mohamed Aymen BOUYAHIA",
    author_email="mohamed-aymen.bouyahia@ensta-paris.fr",
    description="Enhancing LIME through Neural Decision Trees",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/aymen20002005/lime_ndt/tree/main",
    packages=find_packages(),
    install_requires=[
        "numpy",
        "pandas",
        "scikit-learn",
        "matplotlib",
        "keras",
        "tensorflow",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)


# -*- coding: utf-8 -*-
"""
Neural Decision Tree implementation
"""
# import sys

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
# from collections import OrderedDict
from tensorflow.keras import backend as K
from tensorflow.keras import optimizers
from tensorflow.keras.callbacks import Callback
from tensorflow.keras.activations import tanh
from tensorflow.keras.layers import Dense, Input, Layer
from tensorflow.keras.models import Model, Sequential
from tensorflow.keras.regularizers import l1
from tensorflow.keras.utils import to_categorical
from sklearn.metrics import accuracy_score, mean_squared_error, r2_score

# sys.path.append("../utils")
# from common_functions import find_parent
# from common_functions import leaves_id
# from common_functions import get_list_split_phi_forest
import sys
import os
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from utils.common_functions import (get_list_split_phi,
									  get_parents_nodes_leaves_dic,
							  		  print_decision_path)


# from sklearn.preprocessing import LabelEncoder


BIAS = 1
DIM = 0
RIGHT = 1
LEFT = -1
LEAVES = -1
LEAVES_THRESHOLDS = -2
LEAVES_FEATURES = -2
EMPTY_NODE = -5


class SparseNDT(Callback):

	def __init__(self, NDT):
		super().__init__()

		self.weight_masks = []
		weight_matrices = [NDT.W_in_nodes.values, NDT.W_nodes_leaves.values]

		for w in weight_matrices:
			mat = np.copy(w)
			idx = mat.nonzero()
			mat[idx] = 1
			self.weight_masks.append(w)

	def on_batch_end(self, batch, logs=None):
		layers = [2, 5]

		for i, layer in enumerate(layers):
			wb = self.model.layers[layer].get_weights()
			w = wb[0]
			b = wb[1]

			w = w*self.weight_masks[i]

			self.model.layers[layer].set_weights(weights=[w, b])


class tanh_gamma(Layer):

	def __init__(self, gamma=1, **kwargs):
		super(tanh_gamma, self).__init__(**kwargs)
		self.gamma = K.cast_to_floatx(gamma)

	def call(self, inputs):
		return K.tanh(inputs*self.gamma)

	def compute_output_shape(self, input_shape):
		return input_shape


class tanh(Layer):

	def __init__(self, **kwargs):
		super(tanh, self).__init__(**kwargs)

	def call(self, inputs):
		return K.tanh(inputs)

	def compute_output_shape(self, input_shape):
		return input_shape

class CustomEarlyStopping(Callback):
    """
    Custom Early stopping callback that restores the best weights even is there
    was no early stopping during training. (Most od the code is the same as in
    the EarlyStopping callback in Keras)

    Args:
        monitor (str, optional): quantity to monitor. Defaults to 'val_loss'.
        min_delta (float, optional): minimum change in the monitored quantity
            to qualify as an improvement. Defaults to 0.
        patience (float, optional): number of epochs that produced the monitored
            quantity with no improvement after which training will
            be stopped. Defaults to 0.
        verbose (int, optional): verbosity mode. Defaults to 0.
        mode ('auto', 'min', or 'max'; optional): In `min` mode, training will
            stop when the quantity monitored has stopped decreasing; in `max`
            mode it will stop when the quantity monitored has stopped increasing;
            in `auto` mode, the direction is automatically inferred from the name
            of the monitored quantity.. Defaults to 'auto'.
        baseline (float, optional): baseline value for the monitored quantity to reach.
            Training will stop if the model doesn't show improvement over the baseline.
            Defaults to None.
        restore_best_weights (bool, optional): whether to restore model weights from
            the epoch with the best value of the monitored quantity. If False,
            the model weights obtained at the last step of training are used. If
            True, the weghts where the best value of the monitored quantity is achieved
            are restored, even if the training did not stop before the last epoch
            of training. Defaults to False.
    """
    def __init__(self,
                 monitor='val_loss',
                 min_delta=0,
                 patience=0,
                 verbose=0,
                 mode='auto',
                 baseline=None,
                 restore_best_weights=False):

        super(CustomEarlyStopping, self).__init__()

        self.monitor = monitor
        self.patience = patience
        self.verbose = verbose
        self.baseline = baseline
        self.min_delta = abs(min_delta)
        self.wait = 0
        self.stopped_epoch = 0
        self.restore_best_weights = restore_best_weights
        self.best_weights = None
        self.best_weights_epoch = 0

        if mode not in ['auto', 'min', 'max']:
            mode = 'auto'

        if mode == 'min':
            self.monitor_op = np.less
        elif mode == 'max':
            self.monitor_op = np.greater
        else:
            if 'acc' in self.monitor:
                self.monitor_op = np.greater
            else:
                self.monitor_op = np.less

        if self.monitor_op == np.greater:
            self.min_delta *= 1
        else:
            self.min_delta *= -1

    def on_train_begin(self, logs=None):
        # Allow instances to be re-used
        self.wait = 0
        self.stopped_epoch = 0
        self.best_weights_epoch = 0
        if self.baseline is not None:
            self.best = self.baseline
        else:
            self.best = np.inf if self.monitor_op == np.less else -np.inf  # Use np.inf instead of np.Inf

    def on_epoch_end(self, epoch, logs=None):
        current = self.get_monitor_value(logs)
        if current is None:
            return

        if self.monitor_op(current - self.min_delta, self.best):
            self.best = current
            self.wait = 0
            if self.restore_best_weights:  # keep track of the best weights and epoch
                # print(epoch)
                self.best_weights_epoch = epoch
                self.best_weights = self.model.get_weights()
                # print(self.model.get_weights())
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = epoch
                self.model.stop_training = True

    def on_train_end(self, logs=None):
        if self.stopped_epoch > 0 and self.verbose > 0:
            print('Epoch %05d: early stopping' % (self.stopped_epoch + 1))
        if self.restore_best_weights:  # restore best weights even if training did not stop
            if self.verbose > 0:
                print('Restoring model weights from the end of the best epoch:', self.best_weights_epoch)
            self.model.set_weights(self.best_weights)
            # print(self.model.get_weights())

    def get_monitor_value(self, logs):
        logs = logs or {}
        monitor_value = logs.get(self.monitor)
        return monitor_value

class ndt:
	def __init__(self, D, gammas=[10, 1], tree_id=None,
				 sigma=0, gamma_activation=True, is_classifier=False):
		"""
		Create an neural decision tree

		Args:
		gammas (list): Metaparameter for each layer of the neural decision tree
					   (slope of the tanh function).
					   High gamma -> behavior of NN is closer to tree (and also
					   harder to change).
		tree_id (str or int): identifier for the tree.
		sigma (float): STD for the initial weights

		Returns:
		ndt : neural decision tree
		"""
		self.gammas = gammas
		self.use_gamma_activation=gamma_activation
		self.D = D
		self.tree_id = tree_id
		self.sigma = sigma
		self.is_classifier = is_classifier
		if self.is_classifier:
			self._estimator_type = "classifier"
		else:
			self._estimator_type = "regressor"

	def compute_matrices_and_biases(self, decision_tree):
		"""
		Compute synaptic weights and biases according to a decision tree

		Args:
				decision_tree (sklearn.tree.DecisionTreeClassifier): scikit-learn decision
				tree
		"""
		self.decision_tree = decision_tree
		self.splits = pd.DataFrame(get_list_split_phi(decision_tree)).T
		self.leaves = get_parents_nodes_leaves_dic(decision_tree)
		self.N = self.splits.shape[0]
		self.L = len(self.leaves)
		# Fill Input -> Nodes layer matrix
		self.W_in_nodes = pd.DataFrame(np.zeros((self.D, self.N)),
									   index=list(range(self.D)),
									   columns=self.splits.index)
		for node, dim in self.splits[DIM].items():
			self.W_in_nodes.loc[dim, node] = 1.
		# Fill Input -> Nodes layer biases

		self.b_nodes = pd.DataFrame(- self.splits[BIAS])
		self.b_nodes.columns = ["NODES_BIASES"]
		# Fill Nodes -> Leaves layer matrix
		self.W_nodes_leaves = pd.DataFrame(np.zeros((self.N, self.L)),
													 index=self.splits.index,
													 columns=self.leaves.keys())
		for leave, node_sides in self.leaves.items():
			for node, r_l in node_sides:
				self.W_nodes_leaves.loc[node, leave] = r_l
		# Fill Nodes -> Leaves layer biases
		b_leaves = {k: -len(x)+0.5 for k, x in self.leaves.items()}
		self.b_leaves = pd.DataFrame(list(b_leaves.values()),
									 index=b_leaves.keys(),
									 columns=["LEAVES_BIASES"])

		if not self.use_gamma_activation:
			self.W_in_nodes = self.W_in_nodes*self.gammas[0]
			self.b_nodes = self.b_nodes*self.gammas[0]

			self.W_nodes_leaves = self.W_nodes_leaves*self.gammas[1]
			self.b_leaves = self.b_leaves*self.gammas[1]

		if self.is_classifier:
			# Fill Leaves -> class matrix
			self.classes = decision_tree.classes_
			self.C = len(self.classes)

			class_counts_per_leaf = decision_tree.tree_.value[list(self.leaves.keys())]
			class_counts_per_leaf = class_counts_per_leaf.reshape(self.L, self.C)
			class_probas_all_leaves = class_counts_per_leaf * 1/np.sum(class_counts_per_leaf)

   			# self.W_leaves_out = pd.DataFrame(class_counts_per_leaf,
			#                                  index=self.leaves.keys(),
			#                                  columns=self.classes)
			self.W_leaves_out = pd.DataFrame(class_probas_all_leaves,
												index=self.leaves.keys(),
												columns=self.classes)
			# self.W_leaves_out = (self.W_leaves_out.T * 1./self.W_leaves_out.T.sum()).T * 0.5
			self.W_leaves_out = self.W_leaves_out * 0.5
			# self.W_leaves_out *= self.gammas[2]
			# Fill class biases
			# self.b_out = pd.DataFrame(np.zeros(self.C, dtype='double'),
			#                           index=self.classes,
			#                           columns=["CLASS_BIASES"])
			self.b_out = pd.DataFrame(np.sum(class_probas_all_leaves, axis=0),
										index=self.classes,
										columns=["CLASS_BIASES"])
			self.b_out = self.b_out * 0.5


		else:
			self.C = 1
			mean_leaf_values = decision_tree.tree_.value[list(self.leaves.keys())]
			# Debug
			print("mean_leaf_values shape:", mean_leaf_values.shape)
			print("self.L:", self.L, "self.C:", self.C)
			mean_leaf_values = np.squeeze(mean_leaf_values, axis=1)
			print("mean_leaf_values shape after squeeze:", mean_leaf_values.shape)
			self.W_leaves_out = pd.DataFrame(mean_leaf_values[:, :1],  # Ensure shape is (186, 1)
											 index=self.leaves.keys(),
											 columns=["Mean values/1"])
			self.W_leaves_out = self.W_leaves_out / 2

			self.b_out = pd.DataFrame(np.sum(mean_leaf_values),
									  index=[self.C],
									  columns=["BIAS"])
			self.b_out = self.b_out/2

		self.specify_tree_id()

	def specify_tree_id(self):
		"""
		Change the name of the columns/indexes of the weights and biases to  include
		the tree id
		"""
		def add_id_2_elements(list_values, id_to_add):
			append_id = lambda x: str(id_to_add)+"_"+str(x)
			return map(append_id, list_values)
		if self.tree_id is not None:
			# Rename input -> nodes matrix biases
			self.W_in_nodes.columns = add_id_2_elements(self.W_in_nodes.columns,
														self.tree_id)
			self.b_nodes.index = add_id_2_elements(self.b_nodes.index, self.tree_id)
			# Rename nodes -> leaves matrix biases
			self.W_nodes_leaves.index = add_id_2_elements(self.W_nodes_leaves.index,
														  self.tree_id)
			self.W_nodes_leaves.columns = add_id_2_elements(self.W_nodes_leaves.columns,
														  	self.tree_id)
			self.b_leaves.index = add_id_2_elements(self.b_leaves.index, self.tree_id)
			# Rename leaves -> classes matrix biases
			self.W_leaves_out.index = add_id_2_elements(self.W_leaves_out.index,
														self.tree_id)

	def to_keras(self, loss,
				 metrics=[], optimizer=optimizers.Adam,
				 kernel_regularizer=[None, None, None],
				 optimizer_params={"learning_rate": 0.001, "beta_1": 0.9,
								   "beta_2": 0.999, "epsilon": 1e-8,
								   "decay": 1e-6}):
		"""
		Creates a keras neural network

		Args:
				loss (str): loss function
				optimizer (keras.optimizers): keras optimizer
				kernel_regularizer (keras.regularizers): regularization constrains for
														 each layer
				optimizer_params (dict): dictionnary of parameters for the optimizer
		"""

		self.count_ops = 0

		self.input_layer = Input(shape=(self.D,))
		self.nodes_layer = Dense(self.N,
								 kernel_regularizer=kernel_regularizer[0])(self.input_layer)

		self.count_ops = self.count_ops+2*self.D*self.N

		if self.use_gamma_activation:
			self.act_layer_tanh_gamma1 = tanh_gamma(gamma=self.gammas[0])(self.nodes_layer)
			self.count_ops = self.count_ops+self.N*26

		else:
			self.act_layer_tanh_gamma1 = tanh()(self.nodes_layer)
			self.count_ops = self.count_ops+self.N*25

		self.leaves_layer = Dense(self.L,
								  kernel_regularizer=kernel_regularizer[1])(self.act_layer_tanh_gamma1)

		self.count_ops = self.count_ops+2*self.N*self.L

		if self.use_gamma_activation:
			self.act_layer_tanh_gamma2 = tanh_gamma(gamma=self.gammas[1])(self.leaves_layer)
			self.count_ops = self.count_ops+self.L*26

		else:
			self.act_layer_tanh_gamma2 = tanh()(self.leaves_layer)
			self.count_ops = self.count_ops+self.L*25

		if self.is_classifier:
			kr = kernel_regularizer[2]
			self.output_layer = Dense(self.C, activation='softmax',
									  kernel_regularizer=kr)(self.act_layer_tanh_gamma2)

			self.count_ops = self.count_ops+2*self.L*self.C
			self.count_ops = self.count_ops+self.C*(11+self.C*10)

		else:
			kr = kernel_regularizer[2]
			self.output_layer = Dense(self.C,
									  kernel_regularizer=kr)(self.act_layer_tanh_gamma2)

			self.count_ops = self.count_ops+2*self.L*self.C
			self.count_ops = self.count_ops+1

		self.model = Model(inputs=self.input_layer, outputs=self.output_layer)
		self.model_nodes = Model(inputs=self.input_layer, outputs=self.nodes_layer)
		self.model_leaves = Model(inputs=self.input_layer, outputs=self.leaves_layer)

		self.sgd = optimizer(**optimizer_params)
		self.model.compile(loss=loss, optimizer=self.sgd, metrics=metrics)

		# print(self.model.summary())

		flat_b_nodes = self.b_nodes.values.flatten()
		flat_b_leaves = self.b_leaves.values.flatten()
		flat_b_out = self.b_out.values.flatten()

		self.model.layers[1].set_weights(weights=[self.W_in_nodes+np.random.randn(*self.W_in_nodes.shape)*self.sigma,
										 flat_b_nodes+np.random.randn(*flat_b_nodes.shape)*self.sigma])
		self.model.layers[3].set_weights(weights=[self.W_nodes_leaves+np.random.randn(*self.W_nodes_leaves.shape)*self.sigma,
										 flat_b_leaves+np.random.randn(*flat_b_leaves.shape)*self.sigma])
		self.model.layers[5].set_weights(weights=[self.W_leaves_out+np.random.randn(*self.W_leaves_out.shape)*self.sigma,
										 flat_b_out+np.random.randn(*flat_b_out.shape)*self.sigma])

	def fit(self, X, y, sparse=False, epochs=100, min_delta=0, patience=10,
		 	earlyStopping=True, monitor="loss", validation_data=None,
			 restore_best_weights=True, verbose=0, **fit_params):
		"""
		Fit the neural decision tree

		Args:
		X (numpy.array or pandas.DataFrame): Training set
		y (numpy.array or pandas.Series): training set labels
		epochs (int): number of epochs
		min_delta (float): stoping criteria delta
		patience (int): stoping criteria patience
		to_categorical_conversion (bool): If True turn y to categorical
		"""

		callbacks_list = []

		if earlyStopping:
			early_stopping = CustomEarlyStopping(monitor=monitor,
										min_delta=min_delta,
										patience=patience,
										verbose=0,
										mode='auto',
										restore_best_weights=restore_best_weights)

			callbacks_list.append(early_stopping)

		if sparse:

			sparse_ndt = SparseNDT(self)
			callbacks_list.append(sparse_ndt)

		if not callbacks_list:
			callbacks_list = None

		history = self.model.fit(x=X,
								 y=y,
								 callbacks=callbacks_list,
								 epochs=epochs,
								 verbose=verbose,
								 validation_data=validation_data,
								 **fit_params)
		if earlyStopping:
			if early_stopping.stopped_epoch == 0:
				self.stopped_epoch = epochs
			else:
				if restore_best_weights:
					self.stopped_epoch = early_stopping.best_weights_epoch
				else:
					self.stopped_epoch = early_stopping.stopped_epoch


		return history.history

	def get_activations(self,
							X,
							y=None):
		"""
		Get the activation for each layer

		Args:
				X (np.array or pandas.DataFrame): data set
				y (np.array or pandas.Series): labels

		Returns:

		"""
		nodes_a = pd.DataFrame(self.model_nodes.predict(X),
							   columns=self.b_nodes.index)
		leaves_a = pd.DataFrame(self.model_leaves.predict(X),
								columns=self.b_leaves.index)
		output_a = pd.DataFrame(self.model.predict(X),
								columns=self.b_out.index)
		return nodes_a, leaves_a, output_a

	def get_weights_from_NN(self):
		"""
		Get the weights from the keras NN, and load them into attributes of the neural decision tree
		"""
		w_2 = self.model.layers[1].get_weights()

		self.W_in_nodes_nn = pd.DataFrame(w_2[0],
										  index=self.W_in_nodes.index,
										  columns=self.W_in_nodes.columns)
		self.b_nodes_nn = pd.DataFrame(w_2[1],
									   index=self.b_nodes.index,
									   columns=self.b_nodes.columns)

		w_5 = self.model.layers[3].get_weights()
		self.W_nodes_leaves_nn = pd.DataFrame(w_5[0],
											  index=self.W_nodes_leaves.index,
											  columns=self.W_nodes_leaves.columns)
		self.b_leaves_nn = pd.DataFrame(w_5[1],
										index=self.b_leaves.index,
										columns=self.b_leaves.columns)

		w_8 = self.model.layers[5].get_weights()
		self.W_leaves_out_nn = pd.DataFrame(w_8[0],
											index=self.W_leaves_out.index,
											columns=self.W_leaves_out.columns)
		self.b_out_nn = pd.DataFrame(w_8[1],
									   index=self.b_out.index,
									   columns=self.b_out.columns)

	def compute_weights_differences(self):
		"""
		Computes the difference between the original tree weights and those after training
		"""
		self.get_weights_from_NN()
		self.diff_W_in_nodes = self.W_in_nodes - self.W_in_nodes_nn
		self.diff_b_nodes = self.b_nodes - self.b_nodes_nn
		self.diff_W_nodes_leaves = self.W_nodes_leaves - self.W_nodes_leaves_nn
		self.diff_b_leaves = self.b_leaves - self.b_leaves_nn
		self.diff_W_leaves_output = self.W_leaves_out - self.W_leaves_out_nn
		self.diff_b_out = self.b_out - self.b_out_nn

	def predict(self, X, **kwargs):
		return self.model.predict(X, **kwargs)

	def score(self, X, y, **kwargs):
		"""
		Compute prediction score

		Args:
		X (numpy.array or pandas.DataFrame): dataset
		y (numpy.array or pandas.Series): labels
		"""
		if self.is_classifier:
			return accuracy_score(y, self.predict(X, **kwargs))
		else:
			return mean_squared_error(y, self.predict(X, **kwargs))

	def plot_differences(self):
		"""
		Plot the difference between the original tree weights and those after training
		"""
		if "diff_W_in_nodes" not in dir(self):
			self.compute_weights_differences()
		fig = plt.figure(figsize=(3, 2))
		columns = 3
		rows = 2
		ax1a = fig.add_subplot(rows, columns, 1)
		plt.imshow(self.diff_W_in_nodes, aspect="auto", cmap="gray")
		ax1a.set_title("diff W in nodes")

		ax2a = fig.add_subplot(rows, columns, 2)
		plt.imshow(self.diff_b_nodes, aspect="auto", cmap="gray")
		ax2a.set_title("diff b nodes")

		ax3a = fig.add_subplot(rows, columns, 3)
		plt.imshow(self.diff_W_nodes_leaves, aspect="auto", cmap="gray")
		ax3a.set_title("diff W nodes leaves")

		ax4a = fig.add_subplot(rows, columns, 4)
		plt.imshow(self.diff_b_leaves, aspect="auto", cmap="gray")
		ax4a.set_title("diff b leaves")

		ax5a = fig.add_subplot(rows, columns, 5)
		plt.imshow(self.diff_W_leaves_output, aspect="auto", cmap="gray")
		ax5a.set_title("diff W leaves out")

		ax6a = fig.add_subplot(rows, columns, 6)
		plt.imshow(self.diff_b_out, aspect="auto", cmap="gray")
		ax6a.set_title("diff b class")
		plt.show()

	def plot_W_nn_quantiles(self, quantiles=np.arange(0, 99.999, 0.001)):
		"""
		Plot the weights and biases quantiles
		"""
		fig = plt.figure(figsize=(3, 2))
		columns = 3
		rows = 2
		ax1a = fig.add_subplot(rows, columns, 1)
		plt.plot(quantiles, np.percentile(self.W_in_nodes_nn, quantiles))
		plt.plot(quantiles, np.percentile(self.W_in_nodes, quantiles))
		ax1a.set_title("W in nodes")

		ax2a = fig.add_subplot(rows, columns, 2)
		plt.plot(quantiles, np.percentile(self.b_nodes_nn, quantiles))
		plt.plot(quantiles, np.percentile(self.b_nodes, quantiles))
		ax2a.set_title("b nodes")

		ax3a = fig.add_subplot(rows, columns, 3)
		plt.plot(quantiles, np.percentile(self.W_nodes_leaves_nn, quantiles))
		plt.plot(quantiles, np.percentile(self.W_nodes_leaves, quantiles))
		ax3a.set_title("W nodes leaves")

		ax4a = fig.add_subplot(rows, columns, 4)
		plt.plot(quantiles, np.percentile(self.b_leaves_nn, quantiles))
		plt.plot(quantiles, np.percentile(self.b_leaves, quantiles))
		ax4a.set_title("b leaves")

		ax5a = fig.add_subplot(rows, columns, 5)
		plt.plot(quantiles, np.percentile(self.W_leaves_out_nn, quantiles))
		plt.plot(quantiles, np.percentile(self.W_leaves_out, quantiles))
		ax5a.set_title("W leaves out")

		ax6a = fig.add_subplot(rows, columns, 6)
		plt.plot(quantiles, np.percentile(self.b_out_nn, quantiles))
		plt.plot(quantiles, np.percentile(self.b_out, quantiles))
		ax6a.set_title("b class")
		plt.show()

	def neural_network_to_tree(self, node_leaves_matrix=None, in_nodes_matrix=None,
							   nodes_biases=None, threshold=0.9):
		def insert_node(parent,
						right_left,
						children_right,
						children_left,
						node_leaves_matrix,
						threshold):

			print("_____________")
			print("parent", parent)
			print("right_left", right_left)
			print("nodes_leaves_sum")
			print(node_leaves_matrix)

			# Find parent's child
			node_leaves_matrix = node_leaves_matrix[np.abs(node_leaves_matrix).sum(axis=1) > 0]
			count_nb_leaves = np.abs(node_leaves_matrix).sum(axis=1)

			child = count_nb_leaves.argmax()
			print("child", child, count_nb_leaves[child])

			# Remaining nodes are those that have at least one leaf
			remaining_nodes = list(count_nb_leaves.index[count_nb_leaves > 0])

			print("nodes_leaves_sum")
			print(np.abs(node_leaves_matrix).sum(axis=1))
			print("remaining nodes")
			print(remaining_nodes)

			# Add child to tree
			if parent is not None:
				if right_left == RIGHT:
					children_right[parent] = child
				elif right_left == LEFT:
					children_left[parent] = child

			# get leaves of current child
			current_node_leaves = node_leaves_matrix.loc[child]
			print("current node leaves")
			print(current_node_leaves)
			# Remove nodes that have been already included
			if parent in remaining_nodes:
				remaining_nodes.remove(parent)
			if child in remaining_nodes:
				remaining_nodes.remove(child)

			if remaining_nodes:
				node_leaves_matrix = node_leaves_matrix.loc[remaining_nodes]
			leaves = node_leaves_matrix.columns
			right_leaves = leaves[(current_node_leaves >= threshold).values]
			left_leaves = leaves[(current_node_leaves <= -threshold).values]

			print("leaves", leaves)
			print("right_leaves", right_leaves)
			print("left_leaves", left_leaves)

			# Apply the same method to right and left children
			right_node_leaves_matrix = None
			if len(remaining_nodes):
				right_node_leaves_matrix = node_leaves_matrix[right_leaves]
				if np.abs(right_node_leaves_matrix).sum().sum() <= threshold:
					right_node_leaves_matrix = None
				else:
					insert_node(child,
								RIGHT,
								children_right,
								children_left,
								right_node_leaves_matrix,
								threshold)
			if right_node_leaves_matrix is None:
				if not len(right_leaves):
					children_right[child] = EMPTY_NODE
				else:
					children_right[child] = right_leaves[0]
					if len(right_leaves) > 1:
						print("more than one leaf", right_leaves)
					print("adding to", child, "children right", right_leaves)

			left_node_leaves_matrix = None
			if len(remaining_nodes):
				left_node_leaves_matrix = node_leaves_matrix[left_leaves]
				if np.abs(left_node_leaves_matrix).sum().sum() <= threshold:
					left_node_leaves_matrix = None
				else:
					insert_node(child,
								LEFT,
								children_right,
								children_left,
								left_node_leaves_matrix,
								threshold)

			if left_node_leaves_matrix is None:
				if not len(left_leaves):
					children_left[child] = EMPTY_NODE
				else:
					children_left[child] = left_leaves[0]
					if len(left_leaves) > 1:
						print("more than one leaf", left_leaves)
				print("adding to", child, "children left", left_leaves)

		if node_leaves_matrix is None:
			node_leaves_matrix = self.W_nodes_leaves
		if in_nodes_matrix is None:
			in_nodes_matrix = self.W_in_nodes
		if nodes_biases is None:
			nodes_biases = self.b_nodes

		children_right = pd.Series(LEAVES, index=list(node_leaves_matrix.index)+list(node_leaves_matrix.columns))
		children_left = pd.Series(LEAVES, index=list(node_leaves_matrix.index)+list(node_leaves_matrix.columns))
		thresholds = pd.Series(LEAVES_THRESHOLDS, index=list(node_leaves_matrix.index)+list(node_leaves_matrix.columns))
		features = pd.Series(LEAVES_FEATURES, index=list(node_leaves_matrix.index)+list(node_leaves_matrix.columns))
		node_leaves_matrix_local = np.copy(node_leaves_matrix)
		node_leaves_matrix_local = node_leaves_matrix_local * (np.abs(node_leaves_matrix) >= threshold*self.gammas[1])
		insert_node(None,
					None,
					children_right,
					children_left,
					node_leaves_matrix_local,
					threshold)
		# Compute the threshold of each node
		# print features
		# print thresholds
		# print nodes_biases
		thresholds[nodes_biases.index] = nodes_biases["NODES_BIASES"]*1./self.gammas[0]
		# Retrieve the most important feature for each node (one could imagine to create new features if necessary)
		features[in_nodes_matrix.columns] = in_nodes_matrix.index[in_nodes_matrix.values.argmax(axis=0)]
		return children_right, children_left, thresholds, features

	def assert_sample(self, X):
		"""
		Print the decision path for the samples as well as the activation functions.

		Args:
				X (pandas.DataFrame or numpy.array): dataset
		"""
		print_decision_path(self.decision_tree, X)
		print(self.get_activations(X))

	def print_tree_weights(self):
		"""
		Print tree weights
		"""
		print("W: Input -> Nodes")
		print(self.W_in_nodes)
		print("b: Input -> Nodes")
		print(self.b_nodes)
		print("W: Nodes -> Leaves")
		print(self.W_nodes_leaves)
		print("b: Nodes -> Leaves")
		print(self.b_leaves)
		print("W: Leaves -> Out")
		print(self.W_leaves_out)
		print("b: Leaves -> Out")
		print(self.b_out)

	def print_nn_weights(self):
		"""
		Print NN weights
		"""
		print("W: Input -> Nodes")
		print(self.W_in_nodes_nn)
		print("b: Input -> Nodes")
		print(self.b_nodes_nn)
		print("W: Nodes -> Leaves")
		print(self.W_nodes_leaves_nn)
		print("b: Nodes -> Leaves")
		print(self.b_leaves_nn)
		print("W: Leaves -> Out")
		print(self.W_leaves_out_nn)
		print("b: Leaves -> Out")
		print(self.b_out_nn)

	def plot_old_new_network(self):
		"""
		Plot new and old network side by side
		"""
		if "W_in_nodes" not in dir(self):
			self.get_weights_from_NN()
		fig = plt.figure(figsize=(6, 2))
		columns = 6
		rows = 2
		ax1a = fig.add_subplot(rows, columns, 1)
		plt.imshow(self.W_in_nodes, aspect="auto", cmap="gray")
		ax1a.set_title("W in nodes")
		ax1b = fig.add_subplot(rows, columns, 2)
		plt.imshow(self.W_in_nodes_nn, aspect="auto", cmap="gray")
		ax1b.set_title("W in nodes nn")

		ax2a = fig.add_subplot(rows, columns, 3)
		plt.imshow(self.b_nodes, aspect="auto", cmap="gray")
		ax2a.set_title("b nodes ")
		ax2b = fig.add_subplot(rows, columns, 4)
		plt.imshow(self.b_nodes_nn, aspect="auto", cmap="gray")
		ax2b.set_title("b nodes nn")

		ax3a = fig.add_subplot(rows, columns, 5)
		plt.imshow(self.W_nodes_leaves, aspect="auto", cmap="gray")
		ax3a.set_title("W nodes leaves")
		ax3b = fig.add_subplot(rows, columns, 6)
		plt.imshow(self.W_nodes_leaves_nn, aspect="auto", cmap="gray")
		ax3b.set_title("W nodes leaves nn")

		ax4a = fig.add_subplot(rows, columns, 7)
		plt.imshow(self.b_leaves, aspect="auto", cmap="gray")
		ax4a.set_title("b leaves")
		ax4b = fig.add_subplot(rows, columns, 8)
		plt.imshow(self.b_leaves_nn, aspect="auto", cmap="gray")
		ax4b.set_title("b leaves nn")

		ax5a = fig.add_subplot(rows, columns, 9)
		plt.imshow(self.W_leaves_out, aspect="auto", cmap="gray")
		ax5a.set_title("W leaves out")
		ax5b = fig.add_subplot(rows, columns, 10)
		plt.imshow(self.W_leaves_out_nn, aspect="auto", cmap="gray")
		ax5b.set_title("W leaves out nn")

		ax6a = fig.add_subplot(rows, columns, 11)
		plt.imshow(self.b_out, aspect="auto", cmap="gray")
		ax6a.set_title("b class")
		ax6b = fig.add_subplot(rows, columns, 12)
		plt.imshow(self.b_out_nn, aspect="auto", cmap="gray")
		ax6b.set_title("b class nn")
		plt.show()


class NDTClassifier(ndt):
	def __init__(self, D, gammas=[10, 1], tree_id=None, sigma=0, gamma_activation=True):
		super().__init__(D, gammas=gammas, tree_id=tree_id,
						 sigma=sigma, gamma_activation=gamma_activation,
						 is_classifier=True)

	def fit(self, X, y, to_categorical_conversion=True, sparse=False, epochs=100,
			min_delta=0, patience=10, earlyStopping=True,
			monitor='loss', validation_data=None, verbose=0,
			**fit_params):

		if to_categorical_conversion:
			self.to_categorical = True
			y = to_categorical(y)
			if validation_data is not None:
				yv = validation_data[1]
				yv = to_categorical(yv)
				validation_data = (validation_data[0], yv)

		return super().fit(X, y, sparse=sparse, epochs=epochs, min_delta=min_delta,
						   patience=patience, earlyStopping=earlyStopping, monitor=monitor,
						   validation_data=validation_data, verbose=verbose, **fit_params)

	def predict_proba(self, X, **kwargs):
		return super().predict(X, **kwargs)

	def predict_classes(self, X, **kwargs):
		"""
		Predict class membership with the neural decision tree

		Args:
		X (numpy.array or pandas.DataFrame): dataset
		y (numpy.array or pandas.Series): labels
		"""
		class_indexes = np.argmax(self.predict_proba(X, **kwargs), axis=1)
		return self.classes[class_indexes]

	def predict(self, X, **kwargs):
		return self.predict_classes(X, **kwargs)


class NDTRegressor(ndt):
	def __init__(self, D, gammas=[10, 1], tree_id=None, sigma=0, gamma_activation=True):
		super().__init__(D, gammas=gammas, tree_id=tree_id,
						 sigma=sigma, is_classifier=False)


def generateNNClassifier(arch, acts, input_size, n_classes):

	model = Sequential()
	count_ops = 0

	if acts[0] == 'tanh':
		act_ops0 = 25
	else:
		act_ops0 = 1

	if acts[1] == 'tanh':
		act_ops1 = 25
	else:
		act_ops1 = 1

	for index, size in enumerate(arch):
		if index == 0:
			model.add(Dense(size, input_shape=(input_size,), activation=acts[0]))
			count_ops = count_ops+2*input_size*size
			count_ops = count_ops+act_ops0*size

		else:
			model.add(Dense(size, activation=acts[1]))
			count_ops = count_ops+2*arch[index-1]*size
			count_ops = count_ops+act_ops1*size

	optimizer = optimizers.Adam(lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-8, decay=1e-6)

	count_ops = count_ops+2*arch[-1]*n_classes

	if n_classes == 2:
		model.add(Dense(1, activation='sigmoid'))
		count_ops = count_ops+13*n_classes
		model.compile(optimizer=optimizer, loss='binary_crossentropy', metrics=['accuracy'])
	else:
		# print(n_classes)
		model.add(Dense(n_classes, activation='softmax'))
		count_ops = count_ops+n_classes*(11+n_classes*10)
		model.compile(optimizer=optimizer, loss='categorical_crossentropy',
					  metrics=['sparse_categorical_accuracy'])

	model.count_ops = count_ops

	return model


def generateNNRegressor(arch, acts, input_size, n_classes=None):

	model = Sequential()
	count_ops = 0

	if acts[0] == 'tanh':
		act_ops0 = 25
	else:
		act_ops0 = 1

	if acts[1] == 'tanh':
		act_ops1 = 25
	else:
		act_ops1 = 1

	for index, size in enumerate(arch):
		if index == 0:
			model.add(Dense(size, input_shape=(input_size,), activation=acts[0]))
			count_ops = count_ops+2*input_size*size
			count_ops = count_ops+act_ops0*size

		else:
			model.add(Dense(size, activation=acts[1]))
			count_ops = count_ops+2*arch[index-1]*size
			count_ops = count_ops+act_ops1*size

	optimizer = optimizers.Adam(lr=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-8, decay=1e-6)

	count_ops = count_ops+2*arch[-1]*n_classes

	model.add(Dense(1))
	model.compile(optimizer=optimizer, loss='mean_squared_error')

	model.count_ops = count_ops

	return model


if __name__ == "__main__":
	from sklearn.tree import DecisionTreeClassifier  # , export_graphviz
	# import matplotlib.pyplot as plt
	dataset_length = 10000
	D = 2
	X = np.random.randn(dataset_length, D)*0.1
	X[0:dataset_length//2, 0] += 0.1
	X[0:dataset_length//2, 0] += 0.2
	Y = np.ones(dataset_length)
	Y[0:dataset_length//2] *= 0

	X_test = np.random.randn(dataset_length, D)*0.1
	X_test[0:dataset_length//2, 0] += 0.1
	X_test[0:dataset_length//2, 0] += 0.2
	Y_test = np.ones(dataset_length)
	Y_test[0:dataset_length//2] *= 0
	# Train a Tree
	clf = DecisionTreeClassifier(max_depth=10)
	clf = clf.fit(X, Y)

	a = ndt(D=2, gammas=[1, 100, 100], tree_id=0)
	a.compute_matrices_and_biases(clf)
	a.to_keras(loss='mean_squared_error')
	print("FLOPs before:", a.count_ops)
	a.fit(X, Y)
	print("FLOPs after:", a.count_ops)

	print("scores before training")
	print(a.score(X_test, Y_test))
	print(a.score(X, Y))

	print(clf.score(X_test, Y_test))
	print(clf.score(X, Y))
	errors = a.fit(X, Y, epochs=10)
	print("scores after training")
	print(a.score(X_test, Y_test))
	print(a.score(X, Y))